# RCAIDE/Library/Missions/Common/Initialize/differentials_dimensionless.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE Imports 
from RCAIDE.Framework.Core.Arrays  import atleast_2d_col 

# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Differentials
# ----------------------------------------------------------------------------------------------------------------------
def differentials_dimensionless(segment):
    """
    Initializes dimensionless differential operators for mission segment discretization

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed

    Notes
    -----
    This function sets up the numerical discretization operators used for
    solving differential equations in mission segments. It creates the
    control points and differential/integral operators based on the
    specified discretization method.

    The function performs the following steps:
        1. Gets number of control points and discretization method
        2. Generates control points and operators
        3. Ensures proper dimensioning of arrays
        4. Stores results in segment numerics

    **Required Segment State Variables**

    state.numerics:
        - number_of_control_points : int
            Number of points for discretization
        - discretization_method : function
            Method to generate discretization operators

    **Generated Operators**

    numerics.dimensionless:
        - control_points : array
            Normalized points for evaluation
        - differentiate : array
            Differentiation operator matrix
        - integrate : array
            Integration operator matrix

    **Major Assumptions**
        * Valid discretization method provided
        * Number of control points > 1
        * Method generates consistent operators

    Returns
    -------
    None
        Updates segment state directly

    See Also
    --------
    RCAIDE.Framework.Core.Arrays
    RCAIDE.Framework.Mission.Segments
    """     
    
    # unpack
    numerics              = segment.state.numerics
    N                     = numerics.number_of_control_points
    discretization_method = numerics.discretization_method
    
    # get operators
    x,D,I = discretization_method(N,**numerics)
    x = atleast_2d_col(x)
    
    # pack
    numerics.dimensionless.control_points = x
    numerics.dimensionless.differentiate  = D
    numerics.dimensionless.integrate      = I    
    
    return
 


## @ingroup Library-Plots-Performance-Common
# RCAIDE/Library/Plots/Performance/Common/plot_style.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Data 
 
# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
## @ingroup Library-Plots-Performance-Common
def plot_style():
    """
    Create standardized plotting style parameters for visualizations using RCAIDE's standard conventions.

    Returns
    -------
    plot_parameters : Data
        Style parameters with fields:
            - line_width : int
                Width of plot lines, default 2
            - line_style : str
               Style of plot lines, default '-'
            - marker_size : int
                  Size of plot markers, default 8
            - legend_font_size : int
                  Font size for legends, default 12
            - axis_font_size : int
                  Font size for axis labels, default 14
            - title_font_size : int
                  Font size for plot titles, default 18
            - markers : list
                  Collection of marker styles for distinguishing data series
            - color : str
                  Default line color, default 'black'

    Notes
    -----
    Provides consistent styling across all RCAIDE plots including:
         - Line properties (width, style)
         - Marker properties (size, styles)
         - Text properties (font sizes)
         - Color schemes

    The marker list includes a comprehensive set of matplotlib markers
    for distinguishing multiple data series on the same plot.

    **Definitions**

    'Marker'
        Symbol used to highlight individual data points
    
    'Line Style'
        Pattern used for connecting data points

    See Also
    --------
    RCAIDE.Library.Plots.Common.set_axes : Complementary axis styling
    """

    # Universal Plot Settings  
    plot_parameters                  = Data()
    plot_parameters.line_width       = 2 
    plot_parameters.line_style       = '-'
    plot_parameters.marker_size      = 8
    plot_parameters.legend_font_size = 12
    plot_parameters.axis_font_size   = 14
    plot_parameters.title_font_size  = 18    
    plot_parameters.markers          = ['o', 's', '^', 'X', 'd', 'v', 'P', '>','.', ',', 'o', 'v', '^', '<',\
                                        '>', '1', '2', '3', '4', '8', 's', 'p', '*', 'h'\
                                         , 'H', '+', 'x', 'D', 'd', '|', '_'] 
    plot_parameters.color            = 'black'
    
    return plot_parameters


# find_take_off_weight_given_tofl.py
#
# Created:  Sep 2014, C. Ilario, T. Orra 
# Modified: Jan 2016, E. Botero


# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

from RCAIDE.Library.Methods.Performance.estimate_take_off_field_length import estimate_take_off_field_length

import numpy as np

# ----------------------------------------------------------------------
#  Find Takeoff Weight Given TOFL
# ----------------------------------------------------------------------
def find_take_off_weight_given_tofl(vehicle,analyses,target_tofl,altitude = 0, delta_isa = 0,):
    """
    Estimates the maximum allowable takeoff weight for a given takeoff field length requirement.

    Parameters
    ----------
    vehicle : Vehicle
        The vehicle instance containing:
            - mass_properties.operating_empty : float
                Operating empty weight [kg]
            - mass_properties.max_takeoff : float
                Maximum takeoff weight [kg]
    analyses : Analyses
        Container with atmosphere and aerodynamic analyses
    target_tofl : float or ndarray
        Target takeoff field length(s) [m]
    altitude : float, optional
        Airport altitude [m], default 0
    delta_isa : float, optional
        Temperature offset from ISA conditions [K], default 0

    Returns
    -------
    max_tow : ndarray
        Maximum allowable takeoff weight(s) for given field length(s) [kg]

    Notes
    -----
    Uses an interpolation approach by:
        1. Creating array of possible takeoff weights between OEW and 110% MTOW
        2. Computing TOFL for each weight
        3. Interpolating to find weight that gives target TOFL

    **Major Assumptions**
        * Linear interpolation between computed points is valid
        * Target TOFL is within achievable range
        * Meets assumptions from estimate_take_off_field_length

    **Theory**
    Takeoff field length varies approximately with W/T where:
        * W = aircraft weight
        * T = available thrust

    .. math::
        TOFL \propto \\frac{W}{T}

    See Also
    --------
    RCAIDE.Library.Methods.Performance.estimate_take_off_field_length
    """       

    #unpack
    tow_lower = vehicle.mass_properties.operating_empty
    tow_upper = 1.10 * vehicle.mass_properties.max_takeoff

    #saving initial reference takeoff weight
    tow_ref = vehicle.mass_properties.max_takeoff

    tow_vec = np.linspace(tow_lower,tow_upper,50)
    tofl    = np.zeros_like(tow_vec)

    for id,tow in enumerate(tow_vec):
        vehicle.mass_properties.takeoff = tow
        tofl[id], _ = estimate_take_off_field_length(vehicle,analyses,altitude = 0, delta_isa = 0)

    target_tofl = np.atleast_1d(target_tofl)
    max_tow     = np.zeros_like(target_tofl)

    for id,toflid in enumerate(target_tofl):
        max_tow[id] = np.interp(toflid,tofl,tow_vec)

    #reset the initial takeoff weight
    vehicle.mass_properties.max_takeoff = tow_ref

    return max_tow


# RCAIDE/Library/Methods/Geometry/LOPA/LOPA_functions.py
# 
# 
# Created:  Mar 2025, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  get_seat_y_coords
# ----------------------------------------------------------------------------------------------------------------------  
def get_seat_y_coords(cabin,cabin_class,cabin_class_origin): 

    n      = cabin_class.number_of_seats_abrest
    s_w    = cabin_class.seat_width
    ar_w   = cabin_class.seat_arm_rest_width 
    n      = cabin_class.number_of_seats_abrest 
    s_w    = cabin_class.seat_width
    ar_w   = cabin_class.seat_arm_rest_width  
    a_w    = cabin_class.aile_width      

    # determine number of aisles
    if n > 7:
        n_a = 1
    else:
        n_a = 2
        
    # max cabin class width
    total_arm_rest_width =  ((n_a + 1) + n) * ar_w
    total_seat_width     = n * s_w  
    total_aisle_width    = n_a * a_w
    cabin_class_width    = total_arm_rest_width + total_seat_width + total_aisle_width 
    cabin_class.cabin_class_width  = cabin_class_width
    
    ccw    = cabin_class.cabin_class_width
     
    if n == 1:
        s_y_coord = ccw/ 2 - ar_w - s_w /2 
    elif n == 2:
        y_1 = ccw/ 2 - ar_w - s_w /2 
        y_2 = -y_1
        s_y_coord = np.array([y_1, y_2 ])
    elif n == 3:
        y_1 =  ccw / 2 - s_w /2 - ar_w
        y_2 = -y_1 
        y_3 =  y_1 - s_w - ar_w  
        s_y_coord = np.array([ y_1, y_2, y_3 ])  
    elif n == 4:
        if cabin.wide_body:
            y_1 = ccw / 2 - s_w /2 - ar_w   
            y_2 = -y_1    
            y_3 =  s_w /2 + ar_w
            y_4 = -y_3   
            s_y_coord = np.array([ y_1, y_2, y_3,y_4 ])         
        else:
            y_1 = ccw / 2 - s_w /2 - ar_w   
            y_2 = -y_1    
            y_3 = y_1 - s_w - ar_w    
            y_4 = -y_3   
            s_y_coord = np.array([ y_1, y_2, y_3,y_4 ]) 
        
    elif n == 5: 
        y_1 =  ccw / 2 - s_w /2 - ar_w
        y_2 =  -y_1 
        y_3 =  y_1 - s_w - ar_w 
        y_4 =  -y_3
        y_5 =  y_4 + s_w + ar_w  
        s_y_coord = np.array([ y_1, y_2, y_3,y_4,y_5 ])  

    elif n == 6: 
        y_1 =  ccw / 2 - s_w /2 - ar_w
        y_2 =  -y_1 
        y_3 =  y_1 - s_w - ar_w 
        y_4 =  -y_3 
        y_5 =  y_3 - s_w - ar_w 
        y_6 = -y_5   
        s_y_coord = np.array([ y_1, y_2, y_3,y_4, y_5,y_6 ])  

    elif n == 7: 
        y_1 =  ccw / 2 - s_w /2 - ar_w
        y_2 =  -y_1 
        y_3 =  y_1 - s_w - ar_w 
        y_4 =  -y_3 
        y_5 =  s_w + ar_w  
        y_6 =  -y_5 
        y_7 = 0 
        s_y_coord = np.array([ y_1, y_2, y_3,y_4, y_5,y_6,y_7]) 

    elif n == 8:
        y_1 =  ccw / 2 - s_w /2 - ar_w
        y_2 =  -y_1 
        y_3 =  y_1 - s_w - ar_w 
        y_4 =  -y_3 
        y_5 = s_w /2 + ar_w
        y_6 = -y_5 
        y_7 = y_5 + s_w + ar_w 
        y_8 = -y_7  
        
        s_y_coord = np.array([ y_1, y_2, y_3,y_4, y_5,y_6,y_7,y_8 ]) 

    elif n == 9: 
        y_1 =  ccw / 2 - s_w /2 - ar_w
        y_2 =  -y_1 
        y_3 =  y_1 - s_w - ar_w 
        y_4 =  -y_3 
        y_5 =  y_3 - s_w - ar_w  
        y_6 = -y_5 
        y_7 =  s_w + ar_w  
        y_8 =  -y_7 
        y_9 = 0 
        s_y_coord = np.array([ y_1, y_2, y_3,y_4, y_5,y_6,y_7,y_8,y_9])

    elif n == 10:
        y_1 =  ccw / 2 - s_w /2 - ar_w
        y_2 =  -y_1 
        y_3 =  y_1 - s_w - ar_w 
        y_4 =  -y_3 
        y_5 =  y_3 - s_w - ar_w  
        y_6 = -y_5 
        y_7 = s_w /2 + ar_w
        y_8 = -y_7 
        y_9 = y_7 + s_w + ar_w 
        y_10 = -y_9 
        s_y_coord = np.array([ y_1, y_2, y_3,y_4, y_5,y_6,y_7,y_8,y_9,y_10 ])
    
    cabin.width = 2 * (np.max(s_y_coord) + s_w /2 + ar_w)
    s_y_coord  += cabin_class_origin[1] 
    return s_y_coord , cabin_class_origin

# ----------------------------------------------------------------------------------------------------------------------
#  get_seat_x_coords
# ----------------------------------------------------------------------------------------------------------------------  
def get_seat_x_coords(cabin,cabin_class,cabin_class_origin): 
 
    n_r     = cabin_class.number_of_rows
    s_p     = cabin_class.seat_pitch 
    gl_l    = cabin.galley_lavatory_length 
    A_l     = cabin.type_A_door_length       
    ex_p    = cabin.emergency_exit_seat_pitch       
    gl_loc  = cabin_class.galley_lavatory_percent_x_locations         
    ex_loc  = cabin_class.emergency_exit_percent_x_locations 
    A_loc   = cabin_class.type_A_exit_percent_x_locations  
    
    # place seats
    s_x_coord = np.arange(0, n_r) * (s_p)
    normalized_s_x_coord = s_x_coord / s_x_coord[-1]
    
    # create vector of object types (seat,emergency exit,galley/lav,type-A exit)
    object_type = np.zeros((len(s_x_coord),4)) 
    object_type[:,0] =  1 # assign all seats 
    
    # shift for galley and lavatories 
    for i in range(len(gl_loc)):
        var =  normalized_s_x_coord -gl_loc[i]
        loc = np.argmin(abs(var))
        if loc == (n_r -1):
            new_loc =  s_x_coord[-1]  + gl_l 
            s_x_coord = np.append( s_x_coord,new_loc  )
            object_type = np.vstack((object_type, np.array([0, 0, 1, 0])   ))
        else:
            object_type = np.insert(object_type,loc, np.array([0, 0, 1, 0]) , axis=0)
            s_x_coord[loc:] +=  gl_l
            s_x_coord   = np.insert(s_x_coord, loc, s_x_coord[loc] - gl_l)
            if object_type[loc+1, 0] == 1: 
                s_x_coord[loc+1:] +=  s_p / 2
            
    # shift for type A exit
    for i in range(len(A_loc)):
        var = normalized_s_x_coord -A_loc[i]
        loc = np.argmin(abs(var))
        if loc == (n_r -1):
            new_loc =  s_x_coord[-1]  + A_l 
            s_x_coord = np.append( s_x_coord,new_loc  )
            object_type = np.vstack((object_type, np.array([0, 0, 0, 1])   ))
        else:
            object_type = np.insert(object_type,loc, np.array([0, 0, 0, 1]) , axis=0) 
            s_x_coord[loc:] += A_l
            s_x_coord   = np.insert(s_x_coord, loc, s_x_coord[loc] - A_l)
            #if object_type[loc+1, 0] == 1: 
                #s_x_coord[loc+1:] += s_p / 2 
            
    # shift for emergency rows
    for i in range(len(ex_loc)):
        var =  normalized_s_x_coord -ex_loc[i]
        loc =  np.argmin(abs(var))
        if loc == (n_r-1):
            if object_type[-1,1] == 1:
                object_type[-2,1] = 1
            else:
                object_type[-1,1] = 1 
        else:
            if object_type[loc,1] == 1:
                object_type[loc+1,1] = 1   
                s_x_coord[loc+1:] += (ex_p - s_p)  
            else:
                object_type[loc,1] = 1   
                s_x_coord[loc:] += (ex_p - s_p)  
             
    if object_type[0, 0] == 1:
        offset = s_p / 2
    if object_type[0, 2] == 1:
        offset = gl_l / 2
    if object_type[0, 3] == 1:
        offset = A_l / 2            
    s_x_coord += offset
      
    if object_type[-1, 0] == 1:
        offset_end =   s_p / 2  
    if object_type[-1, 2] == 1:
        offset_end=   gl_l / 2  
    if object_type[-1, 3] == 1: 
        offset_end = A_l / 2
     
    cabin_class.length =  s_x_coord[-1] + offset_end
    cabin.length += s_x_coord[-1] + offset_end
    s_x_coord += cabin_class_origin[0]
    cabin_class_origin[0] = s_x_coord[-1] + offset_end
    return s_x_coord , object_type, cabin_class_origin
 

# ----------------------------------------------------------------------------------------------------------------------
#  update_seat_map_layout_using_cabin_taper
# ----------------------------------------------------------------------------------------------------------------------  
def update_seat_map_layout_using_cabin_taper(seat_data,cabin):
    
    cabin_width = cabin.width
    n_fr        = cabin.nose.fineness_ratio 
    nose_length = cabin.width * n_fr
    t_fr        = cabin.tail.fineness_ratio 
    tail_length = cabin.width * t_fr 
    
    # remove components that fall outside of tapered nose region 
    theta1 = np.arctan(cabin_width/nose_length) 
    n_idxs =  np.where(nose_length > seat_data[:,2])[0]
    removed_indexes =  [] 
    for n_i in  range(len(n_idxs)):
        x0    = seat_data[n_i,2] - seat_data[n_i,5]/2
        x1    = seat_data[n_i,2] + seat_data[n_i,5]/2
        y0    = seat_data[n_i,3] - seat_data[n_i,6]/2
        y1    = seat_data[n_i,3] + seat_data[n_i,6]/2
        
        x_pts = np.array([x0, x1])
        y_pts = np.array([y0, y1])
        
        y_border = max(x_pts * np.tan(theta1))
        
        if np.any( y_pts > y_border):
            removed_indexes.append(n_idxs[n_i])
    seat_data = np.delete(seat_data,(removed_indexes), axis=0)
            
        
    # remove components that fall outside of tapered tail region 
    theta2     = np.arctan(cabin_width/tail_length) 
    tail_start = seat_data[-1,2] -  tail_length 
    t_idxs     = np.where(tail_start < seat_data[:,2])[0]
    removed_indexes =  []
    for t_i in  range(len(t_idxs)):
        x0    = seat_data[t_i,2] - seat_data[t_i,5]/2
        x1    = seat_data[t_i,2] + seat_data[t_i,5]/2
        y0    = seat_data[t_i,3] - seat_data[t_i,6]/2
        y1    = seat_data[t_i,3] + seat_data[t_i,6]/2
        
        x_pts = np.array([x0, x1])
        y_pts = np.array([y0, y1])
        
        y_border = max(x_pts * np.tan(theta2)) 
        if np.any( y_pts > y_border): # or remove_feature == True:
            removed_indexes.append(t_idxs[t_i])
    seat_data = np.delete(seat_data,(removed_indexes), axis=0)
            
    return seat_data


# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
import importlib

from RCAIDE.Framework.Core     import Data
from RCAIDE.Framework.Analyses import Analysis  

# ----------------------------------------------------------------------
#  Analysis
# ---------------------------------------------------------------------- 
class Weights(Analysis):
    """ This is a class that call the functions that computes the weight of 
    an aircraft depending on its configration
    
    Assumptions:
        None

    Source:
        N/A

    Inputs:
        None
        
    Outputs:
        None

    Properties Used:
         N/A
    """
    def __defaults__(self):
        """This sets the default values and methods for the weights analysis.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        N/A
        """           
        self.tag                                     = 'weights' 
        self.method                                  = None
        self.vehicle                                 = None
        self.aircraft_type                           = None
        self.propulsion_architecture                 = None 
        self.settings                                = Data()
        self.settings.update_mass_properties         = True
        self.settings.update_center_of_gravity       = True
        self.settings.update_moment_of_inertia       = True
        self.print_weight_analysis_report            = True
        
    def evaluate(self):
        """Evaluate the weight analysis.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        results 
        """
        #unpack
        vehicle = self.vehicle
        
        if self.aircraft_type ==  None:
            raise Exception('Specify Aircraft Type. Current options are: "Transport", "BWB", "General_Aviation" and "VTOL" ')
        
        try:
            compute_module = importlib.import_module(f"RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.{self.propulsion_architecture}.{self.aircraft_type}.{self.method}.compute_operating_empty_weight")
            if self.print_weight_analysis_report:
                print("\nPerforming Weights Analysis")
                print("--------------------------------------------------------")
                print("Propulsion Architecture:", self.propulsion_architecture)
                print("Aircraft Type          :", self.aircraft_type)
                print("Method                 :", self.method)
                
                if  self.settings.update_mass_properties:
                    print("Aircraft operating empty weight will be overwritten")
                if  self.settings.update_center_of_gravity:
                    print("Aircraft center of gravity location will be overwritten")
                if  self.settings.update_moment_of_inertia:
                    print("Aircraft moment of intertia tensor will be overwritten")  
        except:
            raise Exception('Aircraft Type or Weight Buildup Method do not exist!')
        compute_operating_empty_weight = getattr(compute_module, "compute_operating_empty_weight")
        
        # Call the function
        results = compute_operating_empty_weight(vehicle, self.settings) 
        vehicle.mass_properties.weight_breakdown = results
        
        # updating empty weight
        if self.settings.update_mass_properties:  
            vehicle.mass_properties.operating_empty = results.empty.total
            
        # done!
        return results        


# RCAIDE/Methods/Energy/Propulsors/Converters/Turbojet/__init__.py
# 

"""
Contains methods for analyzing and designing turbojet propulsion systems. This module provides functionality for computing thrust, 
sizing engine cores, and evaluating turbojet performance characteristics.

See Also
--------
RCAIDE.Library.Components.Powertrain.Converters.Ram
RCAIDE.Library.Components.Powertrain.Converters.Combustor
RCAIDE.Library.Components.Powertrain.Converters.Compressor
RCAIDE.Library.Components.Powertrain.Converters.Turbine
RCAIDE.Library.Components.Powertrain.Converters.Expansion_Nozzle
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from .append_turbojet_conditions   import append_turbojet_conditions
from .compute_thurst               import compute_thrust
from .size_core                    import size_core 
from .compute_turbojet_performance import compute_turbojet_performance , reuse_stored_turbojet_data
from .design_turbojet              import design_turbojet


# RCAIDE/Methods/Powertrain/Converters/__init__.py
# 

"""
This module provides functionality for modeling energy conversion components in powertrains. It includes methods for 
various energy converters such as combustors, compressors, motors, fans, engines, nozzles, rotors, turbines, and reformers.

See Also
--------
RCAIDE.Library.Components.Powertrain.Converters
RCAIDE.Library.Methods.Powertrain.Distributors
RCAIDE.Library.Methods.Powertrain.Sources
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from . import Combustor
from . import Compression_Nozzle
from . import Compressor
from . import Motor
from . import Ducted_Fan
from . import Engine
from . import Expansion_Nozzle
from . import Supersonic_Nozzle
from . import Fan 
from . import Ram
from . import Rotor
from . import Generator
from . import Turbine
from . import Reformer
from . import Turboshaft
from . import Turboelectric_Generator


# RCAIDE/Library/Attributes/Materials/__init__.py
# 

"""
This module provides material handling capabilities for RCAIDE. It includes classes for managing different types of materials 
and their physical properties, with implementations for various structural and aerospace materials including:
    - Metals (Aluminum, Steel, Nickel, Titanium, Magnesium)
    - Composites (Unidirectional and Bidirectional Carbon Fiber, Carbon Fiber Honeycomb)
    - Polymers (Epoxy, Polyetherimide, Acrylic)
    - Surface Treatments (Paint)

The Solid class serves as the base class for all material implementations.

See Also
--------
RCAIDE.Library.Attributes.Gases : Related module for gas properties
RCAIDE.Library.Attributes.Liquids : Related module for liquid properties
RCAIDE.Library.Attributes.Cryogens : Related module for cryogenic material properties
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Solid                       import Solid
from .Acrylic                     import Acrylic
from .Aluminum                    import Aluminum
from .Aluminum_Alloy              import Aluminum_Alloy
from .Bidirectional_Carbon_Fiber  import Bidirectional_Carbon_Fiber
from .CrossLinked_Polyethylene    import CrossLinked_Polyethylene
from .Copper                      import Copper
from .Epoxy                       import Epoxy 
from .Nickel                      import Nickel
from .Magnesium                   import Magnesium
from .Carbon_Fiber_Honeycomb      import Carbon_Fiber_Honeycomb
from .Paint                       import Paint
from .Polyetherimide              import Polyetherimide
from .Perfluoroalkoxy             import Perfluoroalkoxy
from .Polytetrafluoroethylene     import Polytetrafluoroethylene
from .Polyimide                   import Polyimide
from .Steel                       import Steel
from .Titanium                    import Titanium
from .Unidirectional_Carbon_Fiber import Unidirectional_Carbon_Fiber


# RCAIDE/Library/Compoments/Powertrain/Sources/Cryogenic_Tanks/Cryogenic_Tank.py
# 
# 
# Created:  Jan 2025, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
from RCAIDE.Library.Components          import Component
from RCAIDE.Library.Methods.Powertrain.Sources.Cryogenic_Tanks.append_cryogenic_tank_conditions import append_cryogenic_tank_conditions 

# ----------------------------------------------------------------------------------------------------------------------
#  Cryogenic Tank
# ---------------------------------------------------------------------------------------------------------------------     
class Cryogenic_Tank(Component):
    """
    Base class for aircraft cryogenic tank implementations
    
    Attributes
    ----------
    tag : str
        Identifier for the cryogenic tank (default: 'cryogenic_tank')
    cryogenic_selector_ratio : float
        Ratio of cryogenic flow allocation (default: 1.0)
        
    mass_properties.empty_mass : float
        Mass of empty tank structure [kg] (default: 0.0)
        
    secondary_cryogenic_flow : float
        Secondary cryogenic flow rate [kg/s] (default: 0.0)
        
    cryogenic : Component, optional
        Cryogenic type stored in tank (default: None)

    Notes
    -----
    The cryogenic tank base class provides common attributes and methods for
    different types of aircraft cryogenic tanks. It handles basic cryogenic storage
    and flow management functionality.
    """
    
    def __defaults__(self):
        """
        Sets default values for cryogenic tank attributes
        """          
        self.tag                         = 'cryogenic_tank'
        self.pressure                    = 1.0 
        self.mass_properties.empty_mass  = 0.0   
        self.croygen_selector_ratio      = 1.0 
        self.secondary_cryogenic_flow    = 0.0 
        self.cryogen                     = None
         

    def append_operating_conditions(self,segment,bus):  
        """
        Append cryogenic tank operating conditions for a flight segment
        
        Parameters
        ----------
        segment : Segment
            Flight segment containing state conditions
        bus : Component
            Connected bus component
        """
        append_cryogenic_tank_conditions(self,segment, bus)  
        return                                          


# RCAIDE/Library/Methods/Powertrain/Converters/Rotor/Performance/BEMT_Hemholtz_Vortex_Theory/BEMT_Helmholtz_performance.py
# 
# Created:  Jul 2024, RCAIDE Team 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Core                              import Data , Units, orientation_product, orientation_transpose  
from RCAIDE.Library.Methods.Aerodynamics.Common.Lift    import compute_airfoil_aerodynamics,compute_inflow_and_tip_loss 

# package imports
import  numpy as  np 
# ---------------------------------------------------------------------------------------------------------------------- 
#  BEMT_Helmholtz_performance
# ----------------------------------------------------------------------------------------------------------------------  
def BEMT_Helmholtz_performance(rotor, conditions):
    """
    Analyzes a general rotor given geometry and operating conditions using
    Blade Element Momentum Theory with a Helmholtz Vortex Wake Prescription.
    
    Parameters
    ----------
    rotor : Data
        Rotor component with the following attributes:
            - number_of_blades : int
                Number of blades on the rotor
            - tip_radius : float
                Tip radius of the rotor [m]
            - hub_radius : float
                Hub radius of the rotor [m]
            - twist_distribution : array_like
                Blade twist distribution [radians]
            - chord_distribution : array_like
                Blade chord distribution [m]
            - sweep_distribution : array_like
                Blade sweep distribution [m]
            - radius_distribution : array_like
                Radial station positions [m]
            - thickness_to_chord : array_like
                Thickness-to-chord ratio at each radial station
            - airfoil_polar_stations : array_like
                Indices of airfoil polars for each radial station
            - airfoils : dict
                Dictionary of airfoil objects
            - number_azimuthal_stations : int
                Number of azimuthal stations for 2D analysis
            - nonuniform_freestream : bool
                Flag for nonuniform freestream velocity
            - use_2d_analysis : bool
                Flag to use 2D (azimuthal) analysis
            - body_to_prop_vel : function
                Function to transform velocity from body to propeller frame
            - orientation_euler_angles : list
                Orientation of the rotor [rad, rad, rad]
    conditions : Data
        Flight conditions with:
            - freestream : Data
                Freestream properties
                    - density : array_like
                        Air density [kg/m³]
                    - dynamic_viscosity : array_like
                        Dynamic viscosity [kg/(m·s)]
                    - speed_of_sound : array_like
                        Speed of sound [m/s]
                    - temperature : array_like
                        Temperature [K]
            - frames : Data
                Reference frames
                    - body : Data
                        Body frame
                        - transform_to_inertial : array_like
                            Rotation matrix from body to inertial frame
                    - inertial : Data
                        Inertial frame
                        - velocity_vector : array_like
                            Velocity vector in inertial frame [m/s]
            - energy : Data
                Energy conditions
                    - converters : dict
                        Converter energy conditions indexed by tag
                        - commanded_thrust_vector_angle : array_like
                            Commanded thrust vector angle [rad]
                        - blade_pitch_command : array_like
                            Blade pitch command [rad]
                        - omega : array_like
                            Angular velocity [rad/s]
                        - throttle : array_like
                            Throttle setting [0-1]
                        - design_flag : bool
                            Flag indicating design condition
    
    Returns
    -------
    None

    Notes
    -----
    This function implements the Blade Element Momentum Theory (BEMT) with a Helmholtz
    Vortex Wake model to analyze rotor performance. It calculates detailed aerodynamic
    properties at each blade element and azimuthal position, accounting for 3D wake effects.
    
    The computation follows these steps:
        1. Extract rotor parameters and operating conditions
        2. Transform velocity from inertial to rotor frame
        3. Set up the blade geometry (radial and azimuthal distributions)
        4. Initialize induced velocities
        5. Include effects of rotor incidence and external velocity fields if specified
        6. Compute wake-induced inflow velocities using the Helmholtz wake model
        7. Calculate aerodynamic forces (lift, drag) at each blade element
        8. Compute circulation, thrust, and torque distributions
        9. Apply tip loss corrections
        10. Calculate integrated performance metrics (thrust, power, efficiency)
        11. Store results in the conditions data structure
    
    **Major Assumptions**
        * The wake is modeled using Helmholtz vortex filaments
        * Blade element theory is used to compute local aerodynamic forces
        * Tip losses are modeled using the Prandtl tip loss function
        * Compressibility effects are accounted for through Mach number corrections
        * Reynolds number effects on airfoil performance are included
    
    **Theory**
    The Helmholtz wake model represents the wake as a system of vortex filaments that
    satisfy Helmholtz's vortex theorems. The induced velocities at each blade element
    are computed by applying the Biot-Savart law to these vortex filaments.
    
    The blade forces are calculated using:
        - Lift: :math:`L = 0.5\\cdot\\rho\\cdot W^2\\cdot c\\cdot Cl`
        - Drag: :math:`D = 0.5\\cdot\\rho\\cdot W^2\\cdot c\\cdot Cd`
        - Circulation: :math:`\\Gamma = 0.5\\cdot W\\cdot c\\cdot Cl`
    
    where:
        - ρ is density
        - W is relative velocity
        - c is chord
        - Cl is lift coefficient
        - Cd is drag coefficient
        - Γ is circulation
    
    The thrust and torque are then computed by resolving these forces perpendicular
    and parallel to the rotor plane, respectively.
    
    References
    ----------
    [1] Drela, M. "Qprop Formulation", MIT AeroAstro, June 2006 http://web.mit.edu/drela/Public/web/qprop/qprop_theory.pdf
    [2] Leishman, Gordon J. Principles of helicopter aerodynamicsCambridge university press, 2006.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.Performance.Blade_Element_Momentum_Theory_Helmholtz_Wake.wake_model
    """
    commanded_TV          = conditions.energy.converters[rotor.tag].commanded_thrust_vector_angle
    pitch_c               = conditions.energy.converters[rotor.tag].blade_pitch_command
    eta                   = conditions.energy.converters[rotor.tag].throttle 
    omega                 = conditions.energy.converters[rotor.tag].omega  
    design_flag           = conditions.energy.converters[rotor.tag].design_flag
    B                     = rotor.number_of_blades
    R                     = rotor.tip_radius
    beta_0                = rotor.twist_distribution
    c                     = rotor.chord_distribution
    sweep                 = rotor.sweep_distribution     
    r_1d                  = rotor.radius_distribution
    tc                    = rotor.thickness_to_chord
    a_loc                 = rotor.airfoil_polar_stations
    airfoils              = rotor.airfoils 
    Na                    = rotor.number_azimuthal_stations
    nonuniform_freestream = rotor.nonuniform_freestream
    use_2d_analysis       = rotor.use_2d_analysis
 
    # Unpack freestream conditions
    rho     = conditions.freestream.density[:,0,None]
    mu      = conditions.freestream.dynamic_viscosity[:,0,None]
    a       = conditions.freestream.speed_of_sound[:,0,None]
    T       = conditions.freestream.temperature[:,0,None]
    Vv      = conditions.frames.inertial.velocity_vector
    nu      = mu/rho
    rho_0   = rho 

    # Number of radial stations and segment control points
    Nr       = len(c)
    ctrl_pts = len(Vv)
    
    # Helpful shorthands
    pi      = np.pi 

    # Velocity in the rotor frame
    T_body2inertial         = conditions.frames.body.transform_to_inertial
    T_inertial2body         = orientation_transpose(T_body2inertial)
    V_body                  = orientation_product(T_inertial2body,Vv)
    body2thrust,orientation = rotor.body_to_prop_vel(commanded_TV) 
    T_body2thrust           = orientation_transpose(np.ones_like(T_body2inertial[:])*body2thrust)
    V_thrust                = orientation_product(T_body2thrust,V_body)

    # Check and correct for hover
    V         = V_thrust[:,0,None]
    V[V==0.0] = 1E-6 

    # Non-dimensional radial distribution and differential radius
    chi           = r_1d/R
    diff_r        = np.diff(r_1d)
    deltar        = np.zeros(len(r_1d))
    deltar[1:-1]  = diff_r[0:-1]/2 + diff_r[1:]/2
    deltar[0]     = diff_r[0]/2
    deltar[-1]    = diff_r[-1]/2

    # Calculating rotational parameters
    omegar   = np.outer(omega,r_1d)
    n        = omega/(2.*pi)   # Rotations per second

    # 2 dimensional radial distribution non dimensionalized
    chi_2d         = np.tile(chi[:, None],(1,Na))
    chi_2d         = np.repeat(chi_2d[None,:,:], ctrl_pts, axis=0)
    r_dim_2d       = np.tile(r_1d[:, None] ,(1,Na))
    r_dim_2d       = np.repeat(r_dim_2d[None,:,:], ctrl_pts, axis=0)
    c_2d           = np.tile(c[:, None] ,(1,Na))
    c_2d           = np.repeat(c_2d[None,:,:], ctrl_pts, axis=0)

    # Azimuthal distribution of stations (in direction of rotation)
    psi            = np.linspace(0,2*pi,Na+1)[:-1]
    psi_2d         = np.tile(np.atleast_2d(psi),(Nr,1))
    psi_2d         = np.repeat(psi_2d[None, :, :], ctrl_pts, axis=0)
 
    total_blade_pitch = beta_0  +  pitch_c
    
    # apply blade sweep to azimuthal position
    if np.any(np.array([sweep])!=0):
        use_2d_analysis     = True
        sweep_2d            = np.repeat(sweep[:, None], (1,Na))
        sweep_offset_angles = np.tan(sweep_2d/r_dim_2d)
        psi_2d             += sweep_offset_angles

    # Starting with uniform freestream
    ua       = 0
    ut       = 0
    ur       = 0

    # Include velocities introduced by rotor incidence angles
    if (np.any(abs(V_thrust[:,1]) >1e-3) or np.any(abs(V_thrust[:,2]) >1e-3)) and use_2d_analysis:

        # y-component of freestream in the propeller cartesian plane
        Vy  = V_thrust[:,1,None,None]
        Vy  = np.repeat(Vy, Nr,axis=1)
        Vy  = np.repeat(Vy, Na,axis=2)

        # z-component of freestream in the propeller cartesian plane
        Vz  = V_thrust[:,2,None,None]
        Vz  = np.repeat(Vz, Nr,axis=1)
        Vz  = np.repeat(Vz, Na,axis=2)

        # compute resulting radial and tangential velocities in polar frame
        utz =  -Vz*np.sin(psi_2d)
        urz =   Vz*np.cos(psi_2d)
        uty =  -Vy*np.cos(psi_2d)
        ury =   Vy*np.sin(psi_2d)

        ut +=  (utz + uty)  # tangential velocity in direction of rotor rotation
        ur +=  (urz + ury)  # radial velocity (positive toward tip)
        ua +=  np.zeros_like(ut) 
    
    # Include external velocities introduced by user
    if nonuniform_freestream:
        use_2d_analysis   = True

        # include additional influences specified at rotor sections, shape=(ctrl_pts,Nr,Na)
        ua += rotor.axial_velocities_2d
        ut += rotor.tangential_velocities_2d
        ur += rotor.radial_velocities_2d

    if use_2d_analysis:
        # make everything 2D with shape (ctrl_pts,Nr,Na)

        # 2-D freestream velocity and omega*r
        V_2d   = V_thrust[:,0,None,None]
        V_2d   = np.repeat(V_2d, Na,axis=2)
        V_2d   = np.repeat(V_2d, Nr,axis=1)
        omegar = (np.repeat(np.outer(omega,r_1d)[:,:,None], Na, axis=2))

        # total velocities
        Ua     = V_2d + ua

        # 2-D blade pitch and radial distributions
        if np.size(pitch_c)>1:
            # control variable is the blade pitch, repeat around azimuth
            beta = np.repeat(total_blade_pitch[:,:,None], Na, axis=2)
        else:
            beta = np.tile(total_blade_pitch[:,:,None],(ctrl_pts,1,Na ))

        r    = np.tile(r_1d[None,:,None], (ctrl_pts, 1, Na))
        c    = np.tile(c[None,:,None], (ctrl_pts, 1, Na))
        deltar = np.tile(deltar[None,:,None], (ctrl_pts, 1, Na))

        # 2-D atmospheric properties
        a   = np.tile(np.atleast_2d(a),(1,Nr))
        a   = np.repeat(a[:, :, None], Na, axis=2)
        nu  = np.tile(np.atleast_2d(nu),(1,Nr))
        nu  = np.repeat(nu[:,  :, None], Na, axis=2)
        rho = np.tile(np.atleast_2d(rho),(1,Nr))
        rho = np.repeat(rho[:,  :, None], Na, axis=2)
        T   = np.tile(np.atleast_2d(T),(1,Nr))
        T   = np.repeat(T[:, :, None], Na, axis=2)

    else:
        # total velocities
        r      = r_1d
        Ua     = np.outer((V + ua),np.ones_like(r))
        beta   = total_blade_pitch

    # Total velocities
    Ut     = omegar - ut
    U      = np.sqrt(Ua*Ua + Ut*Ut + ur*ur)

    #---------------------------------------------------------------------------
    # COMPUTE WAKE-INDUCED INFLOW VELOCITIES AND RESULTING ROTOR PERFORMANCE
    #---------------------------------------------------------------------------
    # pack inputs
    wake_inputs                       = Data()
    wake_inputs.velocity_total        = U
    wake_inputs.velocity_axial        = Ua
    wake_inputs.velocity_tangential   = Ut
    wake_inputs.ctrl_pts              = ctrl_pts
    wake_inputs.Nr                    = Nr
    wake_inputs.Na                    = Na        
    wake_inputs.use_2d_analysis       = use_2d_analysis        
    wake_inputs.twist_distribution    = beta
    wake_inputs.chord_distribution    = c
    wake_inputs.radius_distribution   = r
    wake_inputs.speed_of_sounds       = a
    wake_inputs.dynamic_viscosities   = nu
     
    va, vt = RCAIDE.Library.Methods.Powertrain.Converters.Rotor.Performance.Blade_Element_Momentum_Theory_Helmholtz_Wake.wake_model.evaluate_wake(rotor,wake_inputs,conditions)
    
    # compute new blade velocities
    Wa   = va + Ua
    Wt   = Ut - vt

    lamdaw, F, _ = compute_inflow_and_tip_loss(r,R,Wa,Wt,B)

    # Compute aerodynamic forces based on specified input airfoil or surrogate
    Cl, Cdval, alpha, alpha_disc,Ma,W,Re,Re_disc = compute_airfoil_aerodynamics(beta,c,r,R,B,Wa,Wt,a,nu,airfoils,a_loc,ctrl_pts,Nr,Na,tc,use_2d_analysis) 
    
    # compute HFW circulation at the blade
    Gamma = 0.5*W*c*Cl  

    #---------------------------------------------------------------------------      
    # tip loss correction for velocities, since tip loss correction is only applied to loads in prior BET iteration
    va     = F*va
    vt     = F*vt
    lamdaw = r*(va+Ua)/(R*(Ut-vt))

    # More Cd scaling from Mach from AA241ab notes for turbulent skin friction
    Tw_Tinf     = 1. + 1.78*(Ma*Ma)
    Tp_Tinf     = 1. + 0.035*(Ma*Ma) + 0.45*(Tw_Tinf-1.)
    Tp          = (Tp_Tinf)*T
    Rp_Rinf     = (Tp_Tinf**2.5)*(Tp+110.4)/(T+110.4)
    Cd          = ((1/Tp_Tinf)*(1/Rp_Rinf)**0.2)*Cdval

    epsilon                  = Cd/Cl
    epsilon[epsilon==np.inf] = 10.

    # thrust and torque and their derivatives on the blade.
    blade_T_distribution     = rho*(Gamma*(Wt-epsilon*Wa))*deltar
    blade_Q_distribution     = rho*(Gamma*(Wa+epsilon*Wt)*r)*deltar
    blade_dT_dr              = rho*(Gamma*(Wt-epsilon*Wa))
    blade_dQ_dr              = rho*(Gamma*(Wa+epsilon*Wt)*r)


    if use_2d_analysis:
        blade_T_distribution_2d = blade_T_distribution
        blade_Q_distribution_2d = blade_Q_distribution
        blade_dT_dr_2d          = blade_dT_dr
        blade_dQ_dr_2d          = blade_dQ_dr
        blade_Gamma_2d          = Gamma
        alpha_2d                = alpha

        Va_2d = Wa
        Vt_2d = Wt
        V_disc  = np.sqrt(Va_2d**2 + Vt_2d**2)
        M_disc  = Ma
        
        Va_avg = np.average(Wa, axis=2)      # averaged around the azimuth
        Vt_avg = np.average(Wt, axis=2)      # averaged around the azimuth

        Va_ind_2d  = va
        Vt_ind_2d  = vt
        Vt_ind_avg = np.average(vt, axis=2)
        Va_ind_avg = np.average(va, axis=2)

        # set 1d blade loadings to be the average:
        blade_T_distribution    = np.mean((blade_T_distribution_2d), axis = 2)
        blade_Q_distribution    = np.mean((blade_Q_distribution_2d), axis = 2)
        blade_dT_dr             = np.mean((blade_dT_dr_2d), axis = 2)
        blade_dQ_dr             = np.mean((blade_dQ_dr_2d), axis = 2)

        # compute the hub force / rotor drag distribution along the blade
        dL_2d    = 0.5*rho*c_2d*Cd*omegar**2*deltar
        dD_2d    = 0.5*rho*c_2d*Cl*omegar**2*deltar

        rotor_drag_distribution = np.mean(dL_2d*np.sin(psi_2d) + dD_2d*np.cos(psi_2d),axis=2)

    else:
        Va_2d   = np.repeat(Wa[ :, :, None], Na, axis=2)
        Vt_2d   = np.repeat(Wt[ :, :, None], Na, axis=2)

        blade_T_distribution_2d  = np.repeat(blade_T_distribution[:, :, None], Na, axis=2)
        blade_Q_distribution_2d  = np.repeat(blade_Q_distribution[:, :, None], Na, axis=2)
        blade_dT_dr_2d           = np.repeat(blade_dT_dr[:, :, None], Na, axis=2)
        blade_dQ_dr_2d           = np.repeat(blade_dQ_dr[:, :, None], Na, axis=2)
        blade_Gamma_2d           = np.repeat(Gamma[ :, :, None], Na, axis=2)
        alpha_2d                 = np.repeat(alpha[ :, :, None], Na, axis=2)
        V_disc                   = np.sqrt(Va_2d**2 + Vt_2d**2)
        M_disc                   = np.repeat(Ma[ :, :, None], Na, axis=2)
        
        Vt_avg                  = Wt
        Va_avg                  = Wa
        Vt_ind_avg              = vt
        Va_ind_avg              = va
        Va_ind_2d               = np.repeat(va[ :, :, None], Na, axis=2)
        Vt_ind_2d               = np.repeat(vt[ :, :, None], Na, axis=2)

        # compute the hub force / rotor drag distribution along the blade
        dL    = 0.5*rho*c*Cd*omegar**2*deltar
        dL_2d = np.repeat(dL[:, :, None], Na, axis=2)
        dD    = 0.5*rho*c*Cl*omegar**2*deltar
        dD_2d = np.repeat(dD[:, :, None], Na, axis=2)

        rotor_drag_distribution = np.mean(dL_2d*np.sin(psi_2d) + dD_2d*np.cos(psi_2d),axis=2)

    # forces
    thrust                  = np.atleast_2d((B * np.sum(blade_T_distribution, axis = 1))).T
    torque                  = np.atleast_2d((B * np.sum(blade_Q_distribution, axis = 1))).T
    rotor_drag              = np.atleast_2d((B * np.sum(rotor_drag_distribution, axis=1))).T
    power                   = omega*torque

    # calculate coefficients
    D        = 2*R
    Cq       = torque/(rho_0*(n*n)*(D*D*D*D*D))
    Ct       = thrust/(rho_0*(n*n)*(D*D*D*D))
    Cp       = power/(rho_0*(n*n*n)*(D*D*D*D*D))
    Crd      = rotor_drag/(rho_0*(n*n)*(D*D*D*D))
    etap     = V*thrust/power
    A        = np.pi*(R**2 - rotor.hub_radius**2)
    FoM      = thrust*np.sqrt(thrust/(2*rho_0*A))/power  

    # prevent things from breaking
    Cq[Cq<0]                   = 0.
    Ct[Ct<0]                   = 0.
    Cp[Cp<0]                   = 0.
    thrust[omega<0.0]          = -thrust[omega<0.0]
    thrust[omega==0.0]         = 0.0
    power[omega==0.0]          = 0.0
    torque[omega==0.0]         = 0.0
    rotor_drag[omega==0.0]     = 0.0
    Ct[omega==0.0]             = 0.0
    Cp[omega==0.0]             = 0.0
    etap[omega==0.0]           = 0.
    thrust[eta[:,0]  <=0.0]    = 0.0
    power[eta[:,0]  <=0.0]     = 0.0
    torque[eta[:,0]  <=0.0]    = 0.0  
    power[eta>1.0]             = power[eta>1.0]*eta[eta>1.0]
    thrust[eta[:,0]>1.0,:]     = thrust[eta[:,0]>1.0,:]*eta[eta[:,0]>1.0,:] 

    disc_loading           = thrust/(np.pi*(R**2))
    power_loading          = thrust/(power)

    # Make the thrust a 3D vector
    thrust_prop_frame      = np.zeros((ctrl_pts,3))
    thrust_prop_frame[:,0] = thrust[:,0]
    thrust_vector          = orientation_product(orientation_transpose(T_body2thrust),thrust_prop_frame)
     
    conditions.energy.converters[rotor.tag]  = Data( 
                torque                            = torque,
                thrust                            = thrust_vector,  
                power                             = power, 
                azimuthal_distribution            = psi,  
                design_flag                       = design_flag,             
                rpm                               = omega /Units.rpm ,   
                tip_mach                          = omega * R / conditions.freestream.speed_of_sound, 
                efficiency                        = etap,         
                number_radial_stations            = Nr,
                orientation                       = orientation,  
                number_azimuthal_stations         = Na,
                advance_ratio                     = V/(n*D), 
                disc_radial_distribution          = r_dim_2d,
                speed_of_sound                    = conditions.freestream.speed_of_sound,
                density                           = conditions.freestream.density,
                velocity                          = Vv,
                blade_tangential_induced_velocity = Vt_ind_avg,
                blade_axial_induced_velocity      = Va_ind_avg,
                blade_reynolds_number             = Re,
                blade_effective_angle_of_attack   = alpha,
                disc_reynolds_number              = Re_disc,
                disc_effective_angle_of_attack    = alpha_disc,
                blade_tangential_velocity         = Vt_avg,
                blade_axial_velocity              = Va_avg,
                blade_velocity                    = W,
                blade_Mach_number                 = Ma,
                disc_tangential_induced_velocity  = Vt_ind_2d,
                disc_axial_induced_velocity       = Va_ind_2d,
                disc_tangential_velocity          = Vt_2d,
                disc_axial_velocity               = Va_2d,
                disc_velocity                     = V_disc,
                disc_Mach_number                  = M_disc,
                drag_coefficient                  = Cd,
                lift_coefficient                  = Cl, 
                disc_loading                      = disc_loading, 
                power_loading                     = power_loading,      
                omega                             = omega,
                disc_circulation                  = blade_Gamma_2d,
                blade_dT_dr                       = blade_dT_dr,
                disc_dT_dr                        = blade_dT_dr_2d,
                blade_thrust_distribution         = blade_T_distribution,
                disc_thrust_distribution          = blade_T_distribution_2d, 
                thrust_per_blade                  = thrust/B,
                thrust_coefficient                = Ct,
                disc_azimuthal_distribution       = psi_2d,
                blade_dQ_dr                       = blade_dQ_dr,
                disc_dQ_dr                        = blade_dQ_dr_2d,
                blade_torque_distribution         = blade_Q_distribution,
                disc_torque_distribution          = blade_Q_distribution_2d,
                torque_per_blade                  = torque/B,
                torque_coefficient                = Cq,
                power_coefficient                 = Cp, 
                converged_inflow_ratio            = lamdaw, 
                blade_H_distribution              = rotor_drag_distribution,
                rotor_drag                        = rotor_drag,
                rotor_drag_coefficient            = Crd,
                blade_pitch_command               = pitch_c,
                commanded_thrust_vector_angle     = commanded_TV, 
                figure_of_merit                   = FoM, 
        )  

    return 


# RCAIDE/Methods/Energy/Propulsors/Converters/Rotor/__init__.py
# 

"""
Methods for analyzing and designing rotors in aircraft propulsion systems.

This module provides a comprehensive set of functions for the design, analysis, and
performance evaluation of various types of rotors, including propellers, lift rotors,
and prop rotors. It implements different analysis methods such as Blade Element Momentum
Theory (BEMT) with Helmholtz wake modeling and Actuator Disc Theory.

The module includes functions for:
    - Designing optimal rotor geometries for different applications
    - Computing rotor performance across various flight conditions
    - Analyzing wake interactions and induced velocities
    - Setting up rotor operating conditions for mission analysis

These methods support both conventional and advanced aircraft configurations, including
fixed-wing aircraft, helicopters, multicopters, and hybrid VTOL concepts.

See Also
--------
RCAIDE.Library.Components.Powertrain.Converters.Rotor
RCAIDE.Library.Methods.Aerodynamics.Common.Lift
RCAIDE.Library.Methods.Geometry.Airfoil
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from .                          import Design
from .                          import Performance
from .append_rotor_conditions   import append_rotor_conditions
from .compute_rotor_performance import compute_rotor_performance
from .design_propeller          import design_propeller 
from .design_lift_rotor         import design_lift_rotor
from .design_prop_rotor         import design_prop_rotor




# RCAIDE/Library/Attributes/Propellants/Liquid_Petroleum_Gas.py
# 
#
# Created:  Mar 2024, M. Clarke

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 

from .Propellant import Propellant   

# ---------------------------------------------------------------------------------------------------------------------- 
#  Gaseous_Hydrogen Class
# ----------------------------------------------------------------------------------------------------------------------   
class Liquid_Petroleum_Gas(Propellant):
    """
    A class representing Liquid Petroleum Gas (LPG) fuel properties and composition 
    for propulsion applications. 

    Attributes
    ----------
    tag : str
        Identifier for the propellant ('Liquid_Petroleum_Gas')
    reactant : str
        Oxidizer used for combustion ('O2')
    density : float
        Fuel density in kg/m³ (509.26)
    specific_energy : float
        Specific energy content in J/kg (43.1e6)
    energy_density : float
        Energy density in J/m³ (21949.1e6)
    use_high_fidelity_kinetics_model : bool
        Flag for using detailed chemical kinetics (False)
    fuel_surrogate_chemical_properties : dict
        Simplified chemical composition for surrogate model
            - C3H8 : float
                Propane fraction (0.6)
            - C4H10 : float
                Butane fraction (0.4)
    fuel_chemical_properties : dict
        Detailed chemical composition for high-fidelity model
            - NC10H22 : float
                n-Decane fraction (0.16449)
            - NC12H26 : float
                n-Dodecane fraction (0.34308)
            - NC16H34 : float
                n-Hexadecane fraction (0.10335)
            - IC8H18 : float
                iso-Octane fraction (0.08630)
            - NC7H14 : float
                n-Heptene fraction (0.07945)
            - C6H5C2H5 : float
                Ethylbenzene fraction (0.07348)
            - C6H5C4H9 : float
                Butylbenzene fraction (0.05812)
            - C10H7CH3 : float
                Methylnaphthalene fraction (0.10972)
    global_warming_potential_100 : Data
        100-year global warming potentials
            - CO2 : float
                Carbon dioxide (1)
            - H2O : float
                Water vapor (0.06)
            - SO2 : float
                Sulfur dioxide (-226)
            - NOx : float
                Nitrogen oxides (52)
            - Soot : float
                Particulate matter (1166)
            - Contrails : float
                Contrail formation (11)

    Notes
    -----
    LPG is stored as a liquid under moderate pressure at ambient temperature. The 
    composition can vary seasonally and by region, but typically consists of a 
    propane-butane mixture.

    **Definitions**
    
    'Surrogate Model'
        Simplified two-component representation using propane and butane
    
    'High-Fidelity Model'
        Detailed representation including aromatic and aliphatic hydrocarbons
    
    'Global Warming Potential'
        Relative measure of heat trapped in atmosphere compared to CO2

    **Major Assumptions**
        * Surrogate model captures main combustion characteristics
        * Air composition is standard atmospheric
    """

    def __defaults__(self):
        """This sets the default values. 
    
    Assumptions:
        None
    
    Source:
        None
        """    
        self.tag             = 'Liquid_Petroleum_Gas'
        self.reactant        = 'O2'
        self.density         = 509.26                           # kg/m^3 
        self.specific_energy = 43.1e6                           # J/kg
        self.energy_density  = 21949.1e6                        # J/m^3
        
        self.stoichiometric_fuel_air_ratio = 0         # [-] Stoichiometric Fuel to Air ratio
        self.heat_of_vaporization          = 0         # [J/kg] Heat of vaporization at standard conditions
        self.temperature                   = 0         # [K] Temperature of fuel
        self.pressure                      = 0         # [Pa] Pressure of fuel
        self.fuel_surrogate_S1             = {} # [-] Mole fractions of fuel surrogate species
        self.kinetic_mechanism             = '' # [-] Kinetic mechanism for fuel surrogate species
        self.oxidizer                      = ''  
        
        self.global_warming_potential_100.CO2       = 1     # CO2e/kg  
        self.global_warming_potential_100.H2O       = 0.06  # CO2e/kg  
        self.global_warming_potential_100.SO2       = -226  # CO2e/kg  
        self.global_warming_potential_100.NOx       = 52    # CO2e/kg  
        self.global_warming_potential_100.Soot      = 1166  # CO2e/kg    
        self.global_warming_potential_100.Contrails = 11    # kg/CO2e/km          


# RCAIDE/Framework/Networks/Network.py 
#
# Created:  Mar 2025, M.Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  Imports
# ---------------------------------------------------------------------------------------------------------------------
# RCAIDE Imports
import  RCAIDE 
from RCAIDE.Framework.Mission.Common                      import Residuals 
from RCAIDE.Library.Mission.Common.Unpack_Unknowns.energy import unknowns
from RCAIDE.Library.Methods.Powertrain.Systems.compute_avionics_power_draw                import compute_avionics_power_draw
from RCAIDE.Library.Methods.Powertrain.Systems.compute_payload_power_draw                 import compute_payload_power_draw
from RCAIDE.Library.Methods.Powertrain.Converters.Motor.compute_motor_performance         import *
from RCAIDE.Library.Methods.Powertrain.Converters.Generator.compute_generator_performance import * 
from RCAIDE.Library.Components import Component

# python imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Network
# ---------------------------------------------------------------------------------------------------------------------- 
class Network(Component):  
    """ Generalized Hybrid Energy Network (powertrain) Class capable of creating all derivatives of hybrid
    networks, the conventional fuel network and the all-electric network.
    
                                            GENERIC NETWORK
          .........................................:..........................................                          
          :                        :                                :                        :
    .-------------.         .-------------.                 .-------------.           .-------------.                     
    | propulsor 1 |         | propulsor 2 |                 | propulsor 2 |           | propulsor 3 | 
    '-------------'         '-------------'                 '-------------'           '-------------'            
          ||                       ||                              ||                        ||                                  
          ||   .-------------.     ||                              ||  .-------------.       ||
          ||== | converter 1 |====== electric bus / fuel line =========| converter 2 |=======|| 
               '-------------'                                         '-------------'  
                           
    Attributes
    ----------
    tag : str
        Identifier for the network   
    
    Notes
    -----
    The evaluate function is broken into three sections: Section 1 computes all the forces and moments
    from propulsors regardless of if they are powered by fuel or an electrochemical energy storage system;
    Section 2 computees the perfomrance of any converters on the distrution lines, for example,
    turboshafts, motors, pumps etc; and Section 3 computes the thermal mangement of the system as
    well as energy consumtion of the powertrain. The state of storage devices such as covnentional fuel tanks,
    cryogenic tanks and batteries are also updates. Propulsor groups can be "active" or "inactive" to simulate
    engine out conditions. Energy consumtion from payload and avionics is also modeled 
    
    **Definitions** 
    'Propulsor Group'
        Any single or group of compoments that work together to provide thrust.
    
    See Also
    --------
    RCAIDE.Library.Framework.Networks.Fuel
        Fuel network class 
    RCAIDE.Library.Framework.Networks.Fuel_Cell
        Fuel_Cell network class 
    RCAIDE.Library.Framework.Networks.Electric
        All-Electric network class  
    """      
    
    def __defaults__(self):
        """ This sets the default values for the network to function.
        """        
        self.tag                          = 'network'
        self.propulsors                   = Container()  
        self.busses                       = Container()
        self.coolant_lines                = Container()
        self.fuel_lines                   = Container()
        self.converters                   = Container()
        self.identical_propulsors         = True 
        self.reverse_thrust               = False
        self.wing_mounted                 = True   
        self.system_voltage               = None  
        
    # linking the different network components
    def evaluate(network,state,center_of_gravity):
        """ Computes the performance of the network
        """  
        # unpack   
        conditions           = state.conditions 
        busses               = network.busses 
        fuel_lines           = network.fuel_lines 
        coolant_lines        = network.coolant_lines
        converters           = network.converters 
        total_thrust         = 0. * state.ones_row(3) 
        total_mech_power     = 0. * state.ones_row(1) 
        total_elec_power     = 0. * state.ones_row(1) 
        total_moment         = 0. * state.ones_row(3)  
        fuel_mdot            = 0. * state.ones_row(1)
        total_mdot           = 0. * state.ones_row(1)
        cryogen_mdot         = 0. * state.ones_row(1)  
        reverse_thrust       = network.reverse_thrust 
    
        # ----------------------------------------------------------       
        # Section 1.0 Propulsor Performance 
        # ----------------------------------------------------------
        # 1.1 Fuel Propulsors  
        for fuel_line in fuel_lines: 
            for propulsor_group in fuel_line.assigned_propulsors:
                stored_results_flag  = False
                stored_propulsor_tag = None 
                for propulsor_tag in propulsor_group:
                    propulsor            = network.propulsors[propulsor_tag]
                    if propulsor.active and fuel_line.active:   
                        if network.identical_propulsors == False:
                            # run analysis  
                            T,M,P,P_elec,stored_results_flag,stored_propulsor_tag = propulsor.compute_performance(state, center_of_gravity= center_of_gravity)
                        else:             
                            if stored_results_flag == False: 
                                # run propulsor analysis 
                                T,M,P,P_elec,stored_results_flag,stored_propulsor_tag = propulsor.compute_performance(state,center_of_gravity= center_of_gravity)
                            else:
                                # use previous propulsor results 
                                T,M,P,P_elec = propulsor.reuse_stored_data(state,network,stored_propulsor_tag=stored_propulsor_tag,center_of_gravity= center_of_gravity)
        
                        total_thrust      += T   
                        total_moment      += M   
                        total_mech_power  += P 
                        total_elec_power  += P_elec 
        
                        # compute total mass flow rate 
                        fuel_mdot     += conditions.energy.propulsors[propulsor.tag].fuel_flow_rate
                
        # 1.2 Electric Propulsors         
        for bus in busses:           
            bus_conditions       = state.conditions.energy[bus.tag] 
            avionics             = bus.avionics
            payload              = bus.payload  
    
            # Avionics Power Consumtion 
            compute_avionics_power_draw(avionics,bus,conditions)
    
            # Payload Power 
            compute_payload_power_draw(payload,bus,conditions)
    
            # Bus Voltage 
            bus_voltage = bus.voltage * state.ones_row(1)       
    
            if conditions.energy.recharging:             
                bus.charging_current         = bus.nominal_capacity * bus.charging_c_rate 
                charging_power               = (bus.charging_current*bus_voltage*bus.power_split_ratio) 
                bus_conditions.power_draw   -= charging_power/bus.efficiency
                bus_conditions.current_draw  = -bus_conditions.power_draw/bus.voltage
    
            else:
                for propulsor_group in bus.assigned_propulsors:
                    stored_results_flag  = False
                    stored_propulsor_tag = None   
                    for propulsor_tag in propulsor_group:
                        propulsor =  network.propulsors[propulsor_tag]
                        if propulsor.active and bus.active:       
                            if network.identical_propulsors == False:
                                # run analysis  
                                T,M,P_mech,P_elec,stored_results_flag,stored_propulsor_tag = propulsor.compute_performance(state,center_of_gravity= center_of_gravity)
                            else:             
                                if stored_results_flag == False: 
                                    # run propulsor analysis 
                                    T,M,P_mech,P_elec, stored_results_flag,stored_propulsor_tag = propulsor.compute_performance(state,center_of_gravity= center_of_gravity)
                                else:
                                    # use previous propulsor results 
                                    T,M,P_mech,P_elec  = propulsor.reuse_stored_data(state,network,stored_propulsor_tag=stored_propulsor_tag,center_of_gravity=center_of_gravity)
    
                            total_thrust      += T   
                            total_moment      += M   
                            total_mech_power  += P_mech 
                            total_elec_power  += P_elec 
    
                # compute power from each componemnt 
                bus_conditions.power_draw        += (total_elec_power- state.conditions.energy[bus.tag].regenerative_power*bus_voltage ) * bus.power_split_ratio  /bus.efficiency   
                bus_conditions.current_draw       = bus_conditions.power_draw/bus_voltage 
                total_elec_power                 += bus_conditions.power_draw  
             
        # ------------------------------------------------------------------------------------------------------------------- 
        # Section 2.0 Converters
        # -------------------------------------------------------------------------------------------------------------------  
        # 2.1 Fuel Converters         
        for fuel_line in fuel_lines: 
            if fuel_line.active: 
                for converter_group in fuel_line.assigned_converters:
                    for converter_tag in converter_group:
                        converter =  converters[converter_tag]
                        if converter.active:
                            converter.inverse_calculation = True           
                            if isinstance(converter,RCAIDE.Library.Components.Powertrain.Converters.Turboelectric_Generator): 
                                generator             = converter.generator   
                                state.conditions.energy.converters[generator.tag].outputs.power  =  total_elec_power*(1 - state.conditions.energy.hybrid_power_split_ratio ) 
                                P_mech, P_elec, stored_results_flag,stored_propulsor_tag         = converter.compute_performance(state,fuel_line,bus)  
                                bus_conditions.power_draw   += P_elec/bus.efficiency
                                fuel_mdot                   += conditions.energy.converters[converter.tag].fuel_flow_rate  
                 
                            if isinstance(converter,RCAIDE.Library.Components.Powertrain.Converters.Turboshaft):   
                                state.conditions.energy.converters[converter.tag].power = total_mech_power*(1 - state.conditions.energy.hybrid_power_split_ratio )   
                                P_mech, P_elec,stored_results_flag,stored_propulsor_tag = converter.compute_performance(state)   
                                bus_conditions.power_draw   += P_elec/bus.efficiency
                                fuel_mdot                   += conditions.energy.converters[converter.tag].fuel_flow_rate   
                    
        # 2.1 Electric Converters                            
        for bus in busses: 
            if bus.active == True:
                bus_conditions  = state.conditions.energy[bus.tag]            
                for converter_group in bus.assigned_converters:
                    for converter_tag in converter_group:
                        converter =  converters[converter_tag]
                        if converter.active: 
                            converter.inverse_calculation = True
                            if isinstance(converter,RCAIDE.Library.Components.Powertrain.Converters.DC_Motor) or isinstance(converter,RCAIDE.Library.Components.Powertrain.Converters.PMSM_Motor):  
                                compute_motor_performance(converter,conditions)
                                bus_conditions.power_draw   += conditions.energy.converters[converter.tag].inputs.power/bus.efficiency
                                bus_conditions.current_draw  = bus_conditions.power_draw/bus.voltage                            
                                
                            if isinstance(converter,RCAIDE.Library.Components.Powertrain.Converters.DC_Generator) or isinstance(converter,RCAIDE.Library.Components.Powertrain.Converters.PMSM_Generator):                              
                                compute_generator_performance(converter,conditions) 
                                bus_conditions.power_draw   += conditions.energy.converters[converter.tag].outputs.power/bus.efficiency
                                bus_conditions.current_draw  = bus_conditions.power_draw/bus.voltage                            
                        
        # ----------------------------------------------------------        
        # Section 3.0 Sources
        # ----------------------------------------------------------
        # 3.1 Fuel Sources  
        for fuel_line in fuel_lines:     
            for fuel_tank in fuel_line.fuel_tanks:  
                conditions.energy[fuel_line.tag][fuel_tank.tag].mass_flow_rate  += fuel_tank.fuel_selector_ratio*fuel_mdot + fuel_tank.secondary_fuel_flow
                   
        time               = state.conditions.frames.inertial.time[:,0] 
        delta_t            = np.diff(time)
        total_mdot         += fuel_mdot
        
        # 3.2 Electric Sources   
        for bus in  busses:
            if bus.active: 
                for t_idx in range(state.numerics.number_of_control_points):            
                    stored_results_flag       = False
                    stored_battery_cell_tag   = None
                    
                    # ------------------------------------------------------------------------------------------------------------------- 
                    # 3.1 Batteries
                    # -------------------------------------------------------------------------------------------------------------------                
                    for battery_module in  bus.battery_modules:                   
                        if bus.identical_battery_modules == False:
                            # run analysis  
                            stored_results_flag, stored_battery_cell_tag =  battery_module.energy_calc(state,bus,coolant_lines, t_idx, delta_t)
                        else:             
                            if stored_results_flag == False: 
                                # run battery analysis 
                                stored_results_flag, stored_battery_cell_tag  =  battery_module.energy_calc(state,bus,coolant_lines, t_idx, delta_t)
                            else:
                                # use previous battery results 
                                battery_module.reuse_stored_data(state,bus,stored_results_flag, stored_battery_cell_tag)
                      
                    # ------------------------------------------------------------------------------------------------------------------- 
                    # 3.2 Fuel Cell Stacks
                    # ------------------------------------------------------------------------------------------------------------------- 
                    stored_results_flag       = False   
                    stored_fuel_cell_tag      = None                  
                    for fuel_cell_stack in  bus.fuel_cell_stacks:                   
                        if bus.identical_fuel_cell_stacks == False:
                            # run analysis  
                            stored_results_flag, stored_fuel_cell_tag =  fuel_cell_stack.energy_calc(state,bus,coolant_lines, t_idx, delta_t)
                        else:             
                            if stored_results_flag == False: 
                                # run battery analysis 
                                stored_results_flag, stored_fuel_cell_tag  =  fuel_cell_stack.energy_calc(state,bus,coolant_lines, t_idx, delta_t)
                            else:
                                # use previous battery results 
                                fuel_cell_stack.reuse_stored_data(state,bus,stored_results_flag, stored_fuel_cell_tag)
                             
                        # compute cryogen mass flow rate 
                        fuel_cell_stack_conditions  = state.conditions.energy[bus.tag].fuel_cell_stacks[fuel_cell_stack.tag]                        
                        cryogen_mdot[t_idx]        += fuel_cell_stack_conditions.H2_mass_flow_rate[t_idx]
                        
                        # compute total mass flow rate 
                        total_mdot[t_idx]     += fuel_cell_stack_conditions.H2_mass_flow_rate[t_idx]    
                       
                    # Step 3: Compute bus properties          
                    bus.compute_distributor_conditions(state,t_idx, delta_t)
                    
                    # Step 4 : Battery Thermal Management Calculations                    
                    for coolant_line in coolant_lines:
                        if t_idx != state.numerics.number_of_control_points-1: 
                            for heat_exchanger in coolant_line.heat_exchangers: 
                                heat_exchanger.compute_heat_exchanger_performance(state,bus,coolant_line,delta_t[t_idx],t_idx) 
                            for reservoir in coolant_line.reservoirs:   
                                reservoir.compute_reservior_coolant_temperature(state,coolant_line,delta_t[t_idx],t_idx) 
           
                # Step 5: Determine mass flow from cryogenic tanks 
                for cryogenic_tank in bus.cryogenic_tanks:
                    # Step 5.1: Determine the cumulative flow from each cryogen tank
                    fuel_tank_mdot = cryogenic_tank.croygen_selector_ratio*cryogen_mdot + cryogenic_tank.secondary_cryogenic_flow 
                    
                    # Step 5.2: DStore mass flow results 
                    conditions.energy[bus.tag][cryogenic_tank.tag].mass_flow_rate  = fuel_tank_mdot 
                                 
        if reverse_thrust ==  True:
            total_thrust =  total_thrust * -1     
            total_moment =  total_moment * -1                    
        conditions.energy.thrust_force_vector  = total_thrust
        conditions.energy.power                = total_mech_power 
        conditions.energy.thrust_moment_vector = total_moment 
        conditions.weights.vehicle_mass_rate   = total_mdot  
    
        return
    
    def unpack_unknowns(self,segment):
        """Unpacks the unknowns set in the mission to be available for the mission.
    
        Assumptions:
        N/A
        
        Source:
        N/A
        
        Inputs: 
            segment   - data structure of mission segment [-]
        
        Outputs: 
        
        Properties Used:
        N/A
        """            
         
        unknowns(segment)  
        for network in segment.analyses.energy.vehicle.networks:
            # Fuel unknowns 
            for fuel_line_i, fuel_line in enumerate(network.fuel_lines):    
                if fuel_line.active:
                    for propulsor_group in  fuel_line.assigned_propulsors:
                        propulsor = network.propulsors[propulsor_group[0]]
                        propulsor.unpack_propulsor_unknowns(segment)
                        
            # electric unknowns 
            for bus_i, bus in enumerate(network.busses):    
                if bus.active:
                    for propulsor_group in  bus.assigned_propulsors:
                        propulsor = network.propulsors[propulsor_group[0]]
                        propulsor.unpack_propulsor_unknowns(segment) 
        return    
     
    def residuals(self,segment):
        """ This packs the residuals to be sent to the mission solver.
    
           Assumptions:
           None
    
           Source:
           N/A
    
           Inputs:
           state.conditions.energy:
               motor(s).torque                      [N-m]
               rotor(s).torque                      [N-m] 
           residuals soecific to the battery cell   
           
           Outputs:
           residuals specific to battery cell and network
    
           Properties Used: 
           N/A
       """         
        for network in segment.analyses.energy.vehicle.networks:
            for fuel_line_i, fuel_line in enumerate(network.fuel_lines):    
                if fuel_line.active:
                    for propulsor_group in  fuel_line.assigned_propulsors:
                        propulsor =  network.propulsors[propulsor_group[0]]
                        propulsor.pack_propulsor_residuals(segment) 
            for bus_i, bus in enumerate(network.busses):    
                if bus.active:
                    for propulsor_group in  bus.assigned_propulsors:
                        propulsor =  network.propulsors[propulsor_group[0]]
                        propulsor.pack_propulsor_residuals(segment)   
        return      
    
    def add_unknowns_and_residuals_to_segment(self, segment):
        """ This function sets up the information that the mission needs to run a mission segment using this network 
         
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            segment
            eestimated_throttles           [-]
            estimated_propulsor_group_rpms [-]  
            
            Outputs:
            segment
    
            Properties Used:
            N/A
        """                   
        segment.state.residuals.network = Residuals()
        
        for network in segment.analyses.energy.vehicle.networks:
            for propulsor in network.propulsors: 
                propulsor.append_operating_conditions(segment,segment.state.conditions.energy,segment.state.conditions.noise)     
    
            for converter in network.converters: 
                converter.append_operating_conditions(segment,segment.state.conditions.energy)                 
    
            for fuel_line_i, fuel_line in enumerate(network.fuel_lines):   
                # ------------------------------------------------------------------------------------------------------            
                # Create fuel_line results data structure  
                # ------------------------------------------------------------------------------------------------------   
                segment.state.conditions.energy[fuel_line.tag] = RCAIDE.Framework.Mission.Common.Conditions() 
                segment.state.conditions.noise[fuel_line.tag]  = RCAIDE.Framework.Mission.Common.Conditions()   
                 
                # ------------------------------------------------------------------------------------------------------
                # Assign network-specific  residuals, unknowns and results data structures
                # ------------------------------------------------------------------------------------------------------
                if fuel_line.active:
                    for propulsor_group in  fuel_line.assigned_propulsors:
                        propulsor =  network.propulsors[propulsor_group[0]]
                        propulsor.append_propulsor_unknowns_and_residuals(segment)
                        
                # ------------------------------------------------------------------------------------------------------
                # Assign sub component results data structures
                # ------------------------------------------------------------------------------------------------------  
                for fuel_tank in  fuel_line.fuel_tanks: 
                    fuel_tank.append_operating_conditions(segment,fuel_line) 
    
            # ------------------------------------------------------------------------------------------------------            
            # Create bus results data structure  
            # ------------------------------------------------------------------------------------------------------     
            for bus_i, bus in enumerate(network.busses):   
                # ------------------------------------------------------------------------------------------------------            
                # Create bus results data structure  
                # ------------------------------------------------------------------------------------------------------
                segment.state.conditions.energy[bus.tag] = RCAIDE.Framework.Mission.Common.Conditions() 
                segment.state.conditions.noise[bus.tag]  = RCAIDE.Framework.Mission.Common.Conditions()   
    
                # ------------------------------------------------------------------------------------------------------
                # Assign network-specific  residuals, unknowns and results data structures
                # ------------------------------------------------------------------------------------------------------
                if bus.active:
                    for propulsor_group in  bus.assigned_propulsors:
                        propulsor =  network.propulsors[propulsor_group[0]]
                        propulsor.append_propulsor_unknowns_and_residuals(segment)
                        
                # ------------------------------------------------------------------------------------------------------
                # Assign sub component results data structures
                # ------------------------------------------------------------------------------------------------------ 
                bus.append_operating_conditions(segment)
                for battery_module in  bus.battery_modules: 
                    battery_module.append_operating_conditions(segment,bus) 
    
                for fuel_cell_stack in  bus.fuel_cell_stacks: 
                    fuel_cell_stack.append_operating_conditions(segment,bus)      
                    
                for tag, bus_item in bus.items():  
                    if issubclass(type(bus_item), RCAIDE.Library.Components.Component):
                        bus_item.append_operating_conditions(segment,bus)
         
                for cryogenic_tank in  bus.cryogenic_tanks: 
                    cryogenic_tank.append_operating_conditions(segment,bus)
                                                    
    
            for coolant_line_i, coolant_line in enumerate(network.coolant_lines):  
                # ------------------------------------------------------------------------------------------------------            
                # Create coolant_lines results data structure  
                # ------------------------------------------------------------------------------------------------------
                segment.state.conditions.energy[coolant_line.tag] = RCAIDE.Framework.Mission.Common.Conditions()        
                
                # ------------------------------------------------------------------------------------------------------
                # Assign network-specific  residuals, unknowns and results data structures
                # ------------------------------------------------------------------------------------------------------       
                for battery_module in coolant_line.battery_modules: 
                    for btms in battery_module:
                        btms.append_operating_conditions(segment,coolant_line)
                        
                for heat_exchanger in coolant_line.heat_exchangers: 
                    heat_exchanger.append_operating_conditions(segment, coolant_line)
                        
                for reservoir in coolant_line.reservoirs: 
                    reservoir.append_operating_conditions(segment, coolant_line)                           
    
        # Ensure the mission knows how to pack and unpack the unknowns and residuals
        segment.process.iterate.unknowns.network            = self.unpack_unknowns
        segment.process.iterate.residuals.network           = self.residuals   
        
        return segment
 
# ----------------------------------------------------------------------
#  Component Container
# ---------------------------------------------------------------------- 
class Container(Component.Container):
    """ The Network container class 
    """
    def evaluate(self,state,center_of_gravity):
        """ This is used to evaluate the thrust and moments produced by the network.

            Assumptions:  
                If multiple networks are attached their performances will be summed

            Source:
                None 
        """ 
        for net in self.values(): 
            net.evaluate(state,center_of_gravity)  
        return   

# ----------------------------------------------------------------------
#  Handle Linking
# ----------------------------------------------------------------------
Network.Container = Container


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Transport/operating_empty_weight.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ----------------------------------------------------------------------------------------------------------------------
import RCAIDE
from RCAIDE.Framework.Core import Data ,  Units 
from RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.Common import compute_payload_weight
from RCAIDE.Library.Attributes.Materials.Aluminum import Aluminum
import RCAIDE.Library.Methods.Mass_Properties.Weight_Buildups.Conventional.General_Aviation.FLOPS as FLOPS
# python imports 
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
# Operating Empty Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_operating_empty_weight(vehicle, settings=None):
    """"


        Properties Used:
            N/A

        Also creates system components and assigns weights to them. These are appended to the vehicle.
    """
    
    if settings == None:
        W_factors = Data()
        use_max_fuel_weight = True 
    else:
        use_max_fuel_weight = settings.use_max_fuel_weight 
        
    # Set the factors
    if not hasattr(settings, 'weight_reduction_factors'):
        W_factors              = Data() 
        W_factors.main_wing    = 0.
        W_factors.empennage    = 0.
        W_factors.fuselage     = 0.
        W_factors.structural   = 0.
        W_factors.systems      = 0.
    else:
        W_factors = settings.weight_reduction_factors
        if 'structural' in W_factors and W_factors.structural != 0.:
            print('Overriding individual structural weight factors')
            W_factors.main_wing    = 0.
            W_factors.empennage    = 0.
            W_factors.fuselage     = 0.
            W_factors.systems      = 0.
        else:
            W_factors.structural   = 0.
            W_factors.systems      = 0. 
    
    Wings = RCAIDE.Library.Components.Wings  

    # Check vehicle properties:
    if vehicle.flight_envelope.design_mach_number  == None: # Added design mach number
        raise ValueError("FLOPS requires a design mach number for sizing!")
    
    if not hasattr(vehicle, 'flap_ratio'):
        flap_ratio = 0.33
        for wing in vehicle.wings:
            if isinstance(wing, Wings.Main_Wing):
                wing.flap_ratio = flap_ratio 
                
    ##-------------------------------------------------------------------------------             
    # Payload Weight
    ##-------------------------------------------------------------------------------  
    payload = compute_payload_weight(vehicle)
    
    
    vehicle.payload.passengers                      = RCAIDE.Library.Components.Component()
    vehicle.payload.passengers.tag                  = 'passengers'
    vehicle.payload.passengers.mass_properties.mass = payload.passengers
    
    vehicle.payload.baggage                         = RCAIDE.Library.Components.Component()
    vehicle.payload.baggage.tag                     = 'baggage'
    vehicle.payload.baggage.mass_properties.mass    = payload.baggage
    
    vehicle.payload.cargo                           = RCAIDE.Library.Components.Component() 
    vehicle.payload.cargo.tag                       = 'cargo'   
    vehicle.payload.cargo.mass_properties.mass      = payload.cargo    

    ##-------------------------------------------------------------------------------             
    # Operating Items Weight
    ##------------------------------------------------------------------------------- 
    W_oper = FLOPS.compute_operating_items_weight(vehicle)
    
    ##-------------------------------------------------------------------------------         
    # System Weight
    ##------------------------------------------------------------------------------- 
    W_systems = FLOPS.compute_systems_weight(vehicle)
   
    for item in W_systems.keys():
        W_systems[item] *= (1. - W_factors.systems)
    
    ##-------------------------------------------------------------------------------                 
    # Propulsion Weight 
    ##-------------------------------------------------------------------------------
    output                                      = Data()
    output.empty                                = Data() 
    output.empty.propulsion                     = Data() 
    output.empty.propulsion.total               = 0
    output.empty.propulsion.engines             = 0
    output.empty.propulsion.thrust_reversers    = 0
    output.empty.propulsion.miscellaneous       = 0
    output.empty.propulsion.fuel_system         = 0

    W_energy_network                   = Data()
    W_energy_network.total             = 0
    W_energy_network.W_engine          = 0 
    W_energy_network.W_thrust_reverser = 0 
    W_energy_network.W_engine_controls = 0 
    W_energy_network.W_starter         = 0 
    W_energy_network.W_fuel_system     = 0 
    W_energy_network.W_motors          = 0 
    W_energy_network.W_nacelle         = 0 
    W_energy_network.W_battery         = 0
    W_energy_network.W_motor           = 0
    number_of_engines                  = 0
    number_of_tanks                    = 0
    W_energy_network_cumulative        = 0 

    for network in vehicle.networks: 
        W_energy_network_total   = 0 
    
        # Electric-Powered Propulsors  
        for bus in network.busses: 
            # electrical payload 
            W_systems.W_electrical  += bus.payload.mass_properties.mass * Units.kg
     
            # Avionics Weight 
            W_systems.W_avionics  += bus.avionics.mass_properties.mass      
    
            for battery in bus.battery_modules: 
                W_energy_network_total  += battery.mass_properties.mass * Units.kg
                W_energy_network.W_battery = battery.mass_properties.mass * Units.kg
                
        for propulsor in network.propulsors:
            if 'motor' in propulsor:                           
                W_energy_network.W_motor +=  propulsor.motor.mass_properties.mass
                W_energy_network_total  +=  propulsor.motor.mass_properties.mass
                   
    W_energy_network_cumulative += W_energy_network_total
    
    ##-------------------------------------------------------------------------------                 
    # Pod Weight Weight 
    ##-------------------------------------------------------------------------------         
    WPOD  = 0.0             

    NENG   = number_of_engines
    WTNFA  = W_energy_network.W_engine + W_energy_network.W_thrust_reverser + W_energy_network.W_starter \
            + 0.25 * W_energy_network.W_engine_controls + 0.11 * W_systems.W_instruments + 0.13 * W_systems.W_electrical \
            + 0.13 * W_systems.W_hyd_pnu + 0.25 * W_energy_network.W_fuel_system
    WPOD += WTNFA / np.max([1, NENG]) + W_energy_network.W_nacelle / np.max(
        [1.0, NENG + 1. / 2 * (NENG - 2 * np.floor(NENG / 2.))])
 
    output.empty.propulsion.total               = W_energy_network_cumulative
    output.empty.propulsion.battery             = W_energy_network.W_battery
    output.empty.propulsion.motors              = W_energy_network.W_motor
    output.empty.propulsion.engines             = W_energy_network.W_engine
    output.empty.propulsion.thrust_reversers    = W_energy_network.W_thrust_reverser
    output.empty.propulsion.miscellaneous       = W_energy_network.W_engine_controls + W_energy_network.W_starter
    output.empty.propulsion.fuel_system         = W_energy_network.W_fuel_system

    ##-------------------------------------------------------------------------------                 
    # Wing Weights 
    ##------------------------------------------------------------------------------- 
    Al_rho   = Aluminum().density
    Al_sigma = Aluminum().yield_tensile_strength      
    
    num_main_wings      = 0
    W_main_wing        = 0.0
    W_tail_horizontal  = 0.0
    W_tail_vertical    = 0.0
    for wing in vehicle.wings:
        if isinstance(wing, Wings.Main_Wing):
            num_main_wings += 1
    
    for wing in vehicle.wings:
        if isinstance(wing, Wings.Main_Wing): # Main wing
            complexity = settings.FLOPS.complexity
            W_wing = FLOPS.compute_wing_weight(vehicle, wing, WPOD, complexity, settings, num_main_wings)

            # Apply weight factor
            W_wing = W_wing * (1. - W_factors.main_wing) * (1. - W_factors.structural)
            if np.isnan(W_wing):
                W_wing = 0.
            # Pack and sum
            wing.mass_properties.mass = W_wing
            W_main_wing += W_wing
        if isinstance(wing, Wings.Horizontal_Tail): # Horizontal tail
            W_tail = FLOPS.compute_horizontal_tail_weight(vehicle)
            if type(W_tail) == np.ndarray:
                W_tail = sum(W_tail)
            # Apply weight factor
            W_tail = W_tail * (1. - W_factors.empennage) * (1. - W_factors.structural)
            # Pack and sum
            wing.mass_properties.mass = W_tail
            W_tail_horizontal += W_tail
        if isinstance(wing, Wings.Vertical_Tail): # Vertical tail
            W_tail = FLOPS.compute_vertical_tail_weight(vehicle, wing)
            # Apply weight factor
            W_tail = W_tail * (1. - W_factors.empennage) * (1. - W_factors.structural)
            # Pack and sum
            wing.mass_properties.mass = W_tail
            W_tail_vertical += W_tail
        
    ##-------------------------------------------------------------------------------                 
    # Fuselage 
    ##------------------------------------------------------------------------------- 
    W_fuselage_total = 0
    for fuse in vehicle.fuselages:
        W_fuselage = FLOPS.compute_fuselage_weight(vehicle)
        W_fuselage = W_fuselage * (1. - W_factors.fuselage) * (1. - W_factors.structural)
        fuse.mass_properties.mass = W_fuselage
        W_fuselage_total += W_fuselage
    
    ##-------------------------------------------------------------------------------                 
    # Landing Gear Weight
    ##------------------------------------------------------------------------------- 
    landing_gear = FLOPS.compute_landing_gear_weight(vehicle)
    
    ##-------------------------------------------------------------------------------                 
    # Accumulate Structural Weight
    ##-------------------------------------------------------------------------------   
    output.empty.structural                      = Data()
    output.empty.structural.wings                = W_main_wing +   W_tail_horizontal +  W_tail_vertical 
    output.empty.structural.fuselage             = W_fuselage_total
    output.empty.structural.landing_gear         = landing_gear.main +  landing_gear.nose  
    output.empty.structural.nacelle              = W_energy_network.W_nacelle 
    output.empty.structural.paint                = 0  # TODO reconcile FLOPS paint calculations with Raymer and RCAIDE baseline
    output.empty.structural.total                = output.empty.structural.wings   + output.empty.structural.fuselage + output.empty.structural.landing_gear\
                                                   + output.empty.structural.paint + output.empty.structural.nacelle 

    ##-------------------------------------------------------------------------------                 
    # Accumulate Systems Weight
    ##-------------------------------------------------------------------------------
    output.empty.systems                        = Data()
    output.empty.systems.control_systems        = W_systems.W_flight_control
    output.empty.systems.apu                    = W_systems.W_apu
    output.empty.systems.electrical             = W_systems.W_electrical
    output.empty.systems.avionics               = W_systems.W_avionics
    output.empty.systems.hydraulics             = W_systems.W_hyd_pnu
    output.empty.systems.furnishings            = W_systems.W_furnish
    output.empty.systems.air_conditioner        = W_systems.W_ac + W_systems.W_anti_ice # Anti-ice is sometimes included in ECS
    output.empty.systems.instruments            = W_systems.W_instruments
    output.empty.systems.total                  = output.empty.systems.control_systems + output.empty.systems.apu \
                                                    + output.empty.systems.electrical + output.empty.systems.avionics \
                                                    + output.empty.systems.hydraulics + output.empty.systems.furnishings \
                                                    + output.empty.systems.air_conditioner + output.empty.systems.instruments
 
    output.payload    = payload 
    output.operational_items    = Data()
    output.operational_items    = W_oper 
    output.empty.total          = output.empty.structural.total + output.empty.propulsion.total + output.empty.systems.total 
    output.zero_fuel_weight     = output.empty.total + output.operational_items.total + output.payload.total
    output.max_takeoff          = vehicle.mass_properties.max_takeoff
    total_fuel_weight           = vehicle.mass_properties.max_takeoff - output.zero_fuel_weight
    

    ##-------------------------------------------------------------------------------                 
    # Create and assign miscellaneous components and weights to aircraft
    ##-------------------------------------------------------------------------------
    # Assign landing gear weights to landing gear components
    nose_landing_gear = False
    main_landing_gear = False
    for LG in vehicle.landing_gears:
        if isinstance(LG, RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear): # Main landing gear
            LG.mass_properties.mass = landing_gear.main
            main_landing_gear = True
        elif isinstance(LG, RCAIDE.Library.Components.Landing_Gear.Nose_Landing_Gear): # Nose landing gear
            LG.mass_properties.mass = landing_gear.nose
            nose_landing_gear = True
            
    # Append landing gear components to the vehicle if there are none. 
    if nose_landing_gear == False:
        nose_gear = RCAIDE.Library.Components.Landing_Gear.Nose_Landing_Gear()  
        nose_gear.mass_properties.mass = landing_gear.nose    
        vehicle.landing_gears.append(nose_gear)  
    if main_landing_gear == False:
        main_gear = RCAIDE.Library.Components.Landing_Gear.Main_Landing_Gear()  
        main_gear.mass_properties.mass = landing_gear.main  
        vehicle.landing_gears.append(main_gear) 

    # Create components and assign weights for systems
    control_systems                         = RCAIDE.Library.Components.Component()
    control_systems.tag                     = 'control_systems'  
    control_systems.mass_properties.mass    = output.empty.systems.control_systems
    electrical_systems                      = RCAIDE.Library.Components.Component()
    electrical_systems.tag                  = 'electrical_systems'
    electrical_systems.mass_properties.mass = output.empty.systems.electrical
    furnishings                             = RCAIDE.Library.Components.Component()
    furnishings.tag                         = 'furnishings'
    furnishings.mass_properties.mass        = output.empty.systems.furnishings
    air_conditioner                         = RCAIDE.Library.Components.Component() 
    air_conditioner.tag                     = 'air_conditioner'
    air_conditioner.mass_properties.mass    = output.empty.systems.air_conditioner
    apu                                     = RCAIDE.Library.Components.Component()
    apu.tag                                 = 'apu'
    apu.mass_properties.mass                = output.empty.systems.apu
    hydraulics                              = RCAIDE.Library.Components.Component()
    hydraulics.tag                          = 'hydraulics' 
    hydraulics.mass_properties.mass         = output.empty.systems.hydraulics
    avionics                                = RCAIDE.Library.Components.Powertrain.Systems.Avionics()
    avionics.mass_properties.mass           = output.empty.systems.avionics + output.empty.systems.instruments
    optionals                               = RCAIDE.Library.Components.Component()
    optionals.tag                           = 'optionals'
    optionals                               = RCAIDE.Library.Components.Component()
    optionals.mass_properties.mass          = output.operational_items.misc
    
    # assign components to vehicle
    vehicle.systems.control_systems         = control_systems
    vehicle.systems.electrical_systems      = electrical_systems
    vehicle.systems.avionics                = avionics
    vehicle.systems.furnishings             = furnishings
    vehicle.systems.air_conditioner         = air_conditioner 
    vehicle.systems.apu                     = apu
    vehicle.systems.hydraulics              = hydraulics
    vehicle.systems.optionals               = optionals   

    return output


# RCAIDE/Library/Methods/Powertrain/Converters/Rotor/append_rotor_conditions.py
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_rotor_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_rotor_conditions(rotor, segment, energy_conditions, noise_conditions): 
    """
    Initializes and appends rotor conditions to the energy and noise conditions dictionaries.
    
    Parameters
    ----------
    rotor : Rotor
        The rotor component for which conditions are being initialized.
    segment : Segment
        The mission segment in which the rotor is operating.
    energy_conditions : dict
        Dictionary containing energy-related conditions for all propulsion components.
    noise_conditions : dict
        Dictionary containing noise-related conditions for all propulsion components.
    
    Returns
    -------
    None
        This function modifies the energy_conditions and noise_conditions dictionaries in-place.
    
    Notes
    -----
    This function creates empty Conditions objects for the rotor's energy and noise
    characteristics within the respective dictionaries. These conditions will be populated 
    during the mission analysis process.
    
    The energy conditions include various performance metrics such as:
        - Orientation and thrust vector angle
        - Blade pitch command
        - Torque, thrust, and throttle settings
        - RPM and angular velocity
        - Disc and power loading
        - Tip Mach number
        - Efficiency and figure of merit
        - Power coefficient
    
    All values are initialized as zero or one arrays (as appropriate) with the same
    length as the segment's state vector.
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.compute_rotor_performance
    """
    ones_row    = segment.state.ones_row 
    energy_conditions.converters[rotor.tag]                               = Conditions()   
    energy_conditions.converters[rotor.tag].orientation                   = 0. * ones_row(3) 
    energy_conditions.converters[rotor.tag].design_flag                   = False 
    energy_conditions.converters[rotor.tag].commanded_thrust_vector_angle = 0. * ones_row(1) 
    energy_conditions.converters[rotor.tag].blade_pitch_command           = ones_row(1) * rotor.blade_pitch_command 
    energy_conditions.converters[rotor.tag].torque                        = 0. * ones_row(1)
    energy_conditions.converters[rotor.tag].throttle                      = ones_row(1)
    energy_conditions.converters[rotor.tag].thrust                        = 0. * ones_row(1)
    energy_conditions.converters[rotor.tag].rpm                           = 0. * ones_row(1)
    energy_conditions.converters[rotor.tag].omega                         = 0. * ones_row(1)
    energy_conditions.converters[rotor.tag].disc_loading                  = 0. * ones_row(1)                 
    energy_conditions.converters[rotor.tag].power_loading                 = 0. * ones_row(1)
    energy_conditions.converters[rotor.tag].tip_mach                      = 0. * ones_row(1)
    energy_conditions.converters[rotor.tag].efficiency                    = 0. * ones_row(1)
    energy_conditions.converters[rotor.tag].figure_of_merit               = 0. * ones_row(1)
    energy_conditions.converters[rotor.tag].power_coefficient             = 0. * ones_row(1) 
    noise_conditions.converters[rotor.tag]                                = Conditions() 
    return 



# RCAIDE/Library/Attributes/Cryogens/__init__.py
# 

"""
This module provides cryogenic fuel handling capabilities for RCAIDE. It includes classes for managing different types of cryogenic fuels and their properties, with specific implementations for liquid hydrogen and general cryogenic substances.

See Also
--------
RCAIDE.Library.Attributes.Propellants : Related module for conventional fuel handling
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Cryogen         import Cryogen
from .Liquid_Hydrogen import Liquid_Hydrogen


# RCAIDE/Library/Attributes/Atmospheres/__init__.py 
# 

"""
RCAIDE Atmospheres module provides atmospheric models for various planetary environments.

The module includes base atmospheric modeling capabilities through the Atmosphere class,
and specific implementations for Earth, although other planets can be added as needed.

The atmospheric models provide essential environmental parameters needed for aircraft
performance calculations, including: Pressure, Temperature, Density, Speed of sound, and Dynamic viscosity.

See Also
--------
RCAIDE.Library.Methods.Aerodynamics : Aerodynamic calculations using atmospheric properties
RCAIDE.Library.Methods.Missions : Mission analysis utilizing atmospheric conditions
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Atmosphere import Atmosphere

from . import Earth


# RCAIDE/Methods/Powertrain/Sources/Batteries/Ragone/find_specific_power.py
# 
# 
# Created:  Jul 2023, M. Clarke 
 
# ----------------------------------------------------------------------------------------------------------------------
#  METHOD
# ---------------------------------------------------------------------------------------------------------------------- -Ragone
def find_specific_power(battery, specific_energy):
    """
    Determines specific power from a Ragone curve correlation.
    
    Parameters
    ----------
    battery : Battery
        The battery component with the following attributes:
            - cell.ragone.const_1 : float
                Coefficient in the Ragone curve equation [W/kg]
            - cell.ragone.const_2 : float
                Exponent coefficient in the Ragone curve equation [kg/J]
    specific_energy : float
        Specific energy value for which to calculate the specific power [J/kg]
    
    Returns
    -------
    None
        This function modifies the battery object in-place, setting the following attributes:
            - specific_power : float
                Calculated specific power [W/kg]
            - specific_energy : float
                Input specific energy value [J/kg]
    
    Notes
    -----
    This function calculates the specific power of a battery based on its specific energy
    using a Ragone curve correlation. The Ragone curve describes the trade-off between
    specific energy and specific power in energy storage devices.
    
    The correlation used is:
    
    .. math::
        P_{specific} = C_1 \\cdot 10^{C_2 \\cdot E_{specific}}
    
    where:
      - :math:`P_{specific}` is the specific power [W/kg]
      - :math:`E_{specific}` is the specific energy [J/kg]
      - :math:`C_1` and :math:`C_2` are empirical constants
    
    **Major Assumptions**
        * The Ragone curve can be accurately represented by the exponential equation
        * The correlation is valid across the entire range of specific energy values
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Sources.Batteries.Common.find_specific_energy
    """
    
    const_1                 = battery.cell.ragone.const_1
    const_2                 = battery.cell.ragone.const_2
    specific_power          = const_1*10.**(const_2*specific_energy)
    battery.specific_power  = specific_power
    battery.specific_energy = specific_energy


# RCAIDE/Library/Missions/Segments/Common/Initialize/Frames.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Planet Position
# ----------------------------------------------------------------------------------------------------------------------
def planet_position(segment):
    """
    Initializes the vehicle's planetary position coordinates for mission segment

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed

    Notes
    -----
    This function sets up the initial latitude and longitude coordinates
    of the vehicle relative to the planet. It handles position initialization
    from either previous segment conditions or explicit segment parameters.

    The function follows this priority for position initialization:
        1. Previous segment final position (if initials exist)
        2. Explicit segment coordinates (if specified)
        3. Default coordinates (0,0)

    **Required Segment State Variables**

    If segment.state.initials exists:
        state.initials.conditions.frames.planet:
            - longitude : array
                Previous segment final longitude [rad]
            - latitude : array
                Previous segment final latitude [rad]

    state.conditions.frames.planet:
        - longitude : array
            Current segment longitude array [rad]
        - latitude : array
            Current segment latitude array [rad]

    **Optional Segment Parameters**
    
    segment:
        - longitude : float
            Initial longitude [rad]
        - latitude : float
            Initial latitude [rad]

    **Major Assumptions**
        * Spherical planet model
        * Continuous position tracking when using initials
        * Valid coordinate values
        * Radians for angular measurements

    Returns
    -------
    None
        Updates segment conditions directly

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    RCAIDE.Library.Mission.Common.Initialize.inertial_position
    """        
    
    if segment.state.initials:
        longitude_initial = segment.state.initials.conditions.frames.planet.longitude[-1,0]
        latitude_initial  = segment.state.initials.conditions.frames.planet.latitude[-1,0] 
    elif 'latitude' in segment:
        longitude_initial = segment.longitude
        latitude_initial  = segment.latitude      
    else:
        longitude_initial = 0.0
        latitude_initial  = 0.0

    segment.state.conditions.frames.planet.longitude[:,0] = longitude_initial
    segment.state.conditions.frames.planet.latitude[:,0]  = latitude_initial    

    return 


# RCAIDE/Library/Plots/Geometry/Common/contour_surface_slices.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
import plotly.graph_objects as go   

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------  -Common
def contour_surface_slice(x, y, z, values, color_scale, showscale = False, 
                         colorbar_title = None, colorbar_location = 'right', 
                         colorbar_orientation = 'v'):
    """
    Creates a 3D surface plot with contour coloring for visualization of geometry slices.

    Parameters
    ----------
    x : ndarray
        2D array of x-coordinates defining surface points
        
    y : ndarray
        2D array of y-coordinates defining surface points
        
    z : ndarray
        2D array of z-coordinates defining surface points
        
    values : ndarray
        2D array of values used for surface coloring
        
    color_scale : str
        Color scale specification for surface visualization
        
    showscale : bool, optional
        Flag to display the color scale bar (default: False)
        
    colorbar_title : str, optional
        Title for the color scale bar (default: None)
        
    colorbar_location : str, optional
        Location of color scale bar ('right', 'left', 'top', 'bottom') (default: 'right')
        
    colorbar_orientation : str, optional
        Orientation of color scale bar ('v' for vertical, 'h' for horizontal) (default: 'v')

    Returns
    -------
    surface : plotly.graph_objects.Surface
        Surface object ready for plotting

    Notes
    -----
    Creates a surface visualization with:

        - Continuous color mapping based on values
        - Optional color scale bar
        - Customizable appearance
    
    **Major Assumptions**
    
    * Input arrays are 2D and of matching dimensions
    * Values array corresponds to surface points
    * Color scale is valid for plotly
    """
    return go.Surface(
        x=x, 
        y=y, 
        z=z, 
        surfacecolor=values, 
        colorscale=color_scale, 
        showscale=showscale, 
        colorbar=dict(
            title=dict(text=colorbar_title),  # Use a dictionary for title
            orientation="v"  # Vertical orientation (default)
        )
    )
   



# Nickel.py
#
# Created: Jul, 2017, J. Smart
# Modified: Apr, 2018, J. Smart

#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------

from .Solid import Solid
from RCAIDE.Framework.Core import Units

#-------------------------------------------------------------------------------
# Cold Rolled Nickel/Cobalt Chromoly Alloy Solid Class
#-------------------------------------------------------------------------------

class Nickel(Solid):
    """
    A class representing nickel alloy material properties.

    Attributes
    ----------
    ultimate_tensile_strength : float
        Maximum tensile stress before failure in Pa (1830e6)
    ultimate_shear_strength : float
        Maximum shear stress before failure in Pa (1050e6)
    ultimate_bearing_strength : float
        Maximum bearing stress before failure in Pa (1830e6)
    yield_tensile_strength : float
        Stress at which material begins to deform plastically in Pa (1550e6)
    yield_shear_strength : float
        Shear stress at which material begins to deform plastically in Pa (1050e6)
    yield_bearing_strength : float
        Bearing stress at which material begins to deform plastically in Pa (1550e6)
    minimum_gage_thickness : float
        Minimum manufacturable thickness in m (0.0)
    density : float
        Material density in kg/m³ (8430)

    Notes
    -----
    This class implements material properties for nickel alloy based on 
    median values from manufacturer reported data. The zero value for minimum gage thickness 
    indicates that this should be specified based on specific manufacturing capabilities.

    **Definitions**
    
    'Ultimate Strength'
        The maximum stress that a material can withstand before failure
    
    'Yield Strength'
        The stress at which a material begins to deform plastically

    References
    ----------
    [1] MatWeb. (n.d.).  Overview of materials for Nickel Alloy. Overview of materials for nickel alloy. https://www.matweb.com/search/DataSheet.aspx?MatGUID=8808b026f7c14d2f8d61f2d476aaeb26 
    """
    def __defaults__(self):
        """Sets material properties at instantiation.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        N/A

        Outputs:
        N/A

        Properties Used:
        None
        """

        self.ultimate_tensile_strength  = 1830e6    * Units.Pa
        self.ultimate_shear_strength    = 1050e6    * Units.Pa
        self.ultimate_bearing_strength  = 1830e6    * Units.Pa
        self.yield_tensile_strength     = 1550e6    * Units.Pa
        self.yield_shear_strength       = 1050e6    * Units.Pa
        self.yield_bearing_strength     = 1550e6    * Units.Pa
        self.minimum_gage_thickness     = 0.0       * Units.m
        self.density                    = 8430.     * Units['kg/(m**3)']



## @ingroup Library-Plots-Performance-Common
# RCAIDE/Library/Plots/Performance/Common/set_axes.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------    
## @ingroup Library-Plots-Performance-Common
def set_axes(axes):
    """
    Apply a standardized RCAIDE formatting style to matplotlib plot axes.

    Parameters
    ----------
    axes : matplotlib.axes.Axes
        Axes object to be formatted with RCAIDE standard style

    Returns
    -------
    None

    Notes
    -----
    Applies the following formatting:
        - Minor tick marks enabled
        - Major grid lines: solid grey, width 0.5
        - Minor grid lines: dotted grey, width 0.5
        - Grid lines enabled
        - Scientific notation disabled for y-axis
        - Axis offset disabled for y-axis

    This function ensures consistent axis appearance across all RCAIDE plots
    and should be called after plotting data but before displaying the figure.

    **Definitions**

    'Major Grid'
        Primary grid lines at major tick marks
    
    'Minor Grid'
        Secondary grid lines at minor tick marks
    
    'Tick Marks'
        Small lines indicating axis scale divisions

    See Also
    --------
    RCAIDE.Library.Plots.Common.plot_style : Complementary plot styling
    """

    axes.minorticks_on()
    axes.grid(which='major', linestyle='-', linewidth=0.5, color='grey')
    axes.grid(which='minor', linestyle=':', linewidth=0.5, color='grey')
    axes.grid(True)
    axes.get_yaxis().get_major_formatter().set_scientific(False)
    axes.get_yaxis().get_major_formatter().set_useOffset(False)

    return



# RCAIDE/Methods/Powertrain/Sources/Batteries/Lithium_Ion_NMC/compute_nmc_cell_performance.py
# 
# 
# Created:  Feb 2024, M. Clarke
# Modified: Sep 2024, S. Shekar

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Framework.Core                       import Units 
import numpy as np
from copy import  deepcopy
 
# ----------------------------------------------------------------------------------------------------------------------
# compute_nmc_cell_performance
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_nmc_cell_performance(battery_module, state, bus, coolant_lines, t_idx, delta_t):
    """
    Computes the performance of a lithium-nickel-manganese-cobalt-oxide (NMC) battery cell.

    Parameters
    ----------
    battery_module : RCAIDE.Library.Components.Sources.Battery_Modules.Lithium_Ion_NMC
        Battery module component with the following attributes:
            - tag : str
                Identifier for the battery module
            - cell : Data
                Cell properties
                    - electrode_area : float
                        Area of the electrode [m²]
                    - surface_area : float
                        Surface area of the cell [m²]
                    - mass : float
                        Mass of a single cell [kg]
                    - specific_heat_capacity : float
                        Specific heat capacity of the cell [J/(kg·K)]
                    - discharge_performance_map : function
                        Function that maps state of charge, temperature, and current to voltage
            - maximum_energy : float
                Maximum energy capacity of the module [J]
            - electrical_configuration : Data
                Electrical configuration
                    - series : int
                        Number of cells in series
                    - parallel : int
                        Number of cells in parallel
    state : RCAIDE.Framework.Mission.Common.State
        State object containing:
            - conditions : Data
                Flight conditions
                    - energy : dict
                        Energy conditions indexed by component tag
                            - [bus.tag] : Data
                                Bus-specific conditions
                                    - energy : numpy.ndarray
                                        Energy stored in the bus [J]
                                    - power_draw : numpy.ndarray
                                        Power draw on the bus [W]
                                    - current_draw : numpy.ndarray
                                        Current draw on the bus [A]
                                    - battery_modules : dict
                                        Battery module conditions indexed by tag
                                            - [battery_module.tag] : Data
                                                Battery module conditions
                                                    - energy : numpy.ndarray
                                                        Energy stored in the module [J]
                                                    - voltage_open_circuit : numpy.ndarray
                                                        Open-circuit voltage [V]
                                                    - power : numpy.ndarray
                                                        Power output [W]
                                                    - internal_resistance : numpy.ndarray
                                                        Internal resistance [Ω]
                                                    - heat_energy_generated : numpy.ndarray
                                                        Heat energy generated [W]
                                                    - voltage_under_load : numpy.ndarray
                                                        Voltage under load [V]
                                                    - current : numpy.ndarray
                                                        Current [A]
                                                    - temperature : numpy.ndarray
                                                        Temperature [K]
                                                    - state_of_charge : numpy.ndarray
                                                        State of charge [0-1]
                                                    - cell : Data
                                                        Cell-specific conditions with same properties as module
            - numerics : Data
                Numerical properties
                    - number_of_control_points : int
                        Number of control points in the mission
    bus : RCAIDE.Library.Components.Systems.Electrical_Bus
        Electrical bus component with the following attributes:
            - tag : str
                Identifier for the electrical bus
            - battery_module_electric_configuration : str
                Configuration of battery modules ("Series" or "Parallel")
            - battery_modules : list
                List of battery modules connected to the bus
    coolant_lines : list
        List of coolant lines for thermal management
    t_idx : int
        Current time index in the simulation
    delta_t : numpy.ndarray
        Time step size [s]

    Returns
    -------
    stored_results_flag : bool
        Flag indicating if results were stored
    stored_battery_module_tag : str
        Tag of the battery module for which results were stored

    Notes
    -----
    This function models the electrical and thermal behavior of an NMC battery cell
    based on experimental data. It updates various battery conditions in the `state` object,
    including: current energy, temperature, heat energy generated, load power, current,
    open-circuit voltage, charge throughput, internal resistance, state of charge,
    depth of discharge, and voltage under load.

    The model includes:
        - Internal resistance calculation
        - Thermal modeling (heat generation and temperature change)
        - Electrical performance (voltage and current calculations)
        - State of charge and depth of discharge updates

    **Major Assumptions**
        * All battery modules exhibit the same thermal behavior
        * The cell temperature is assumed to be the temperature of the entire module
        * Battery performance follows empirical models based on experimental data

    **Theory**
    The internal resistance is modeled as a function of state of charge:
    
    .. math::
        R_0 = 0.01483 \\cdot SOC^2 - 0.02518 \\cdot SOC + 0.1036
    
    Heat generation includes both Joule heating and entropy effects:
    
    .. math::
        \\dot{q}_{entropy} = -T \\cdot \\Delta S \\cdot i / (nF)
        
        \\dot{q}_{joule} = i^2 / \\sigma
        
        Q_{heat} = (\\dot{q}_{joule} + \\dot{q}_{entropy}) \\cdot A_s

    References
    ----------
    [1] Zou, Y., Hu, X., Ma, H., and Li, S. E., "Combined State of Charge and State of Health estimation over lithium-ion battery cell cycle lifespan for electric vehicles," Journal of Power Sources, Vol. 273, 2015, pp. 793-803. doi:10.1016/j.jpowsour.2014.09.146
    [2] Jeon, D. H., and Baek, S. M., "Thermal modeling of cylindrical lithium ion battery during discharge cycle," Energy Conversion and Management, Vol. 52, No. 8-9, 2011, pp. 2973-2981. doi:10.1016/j.enconman.2011.04.013

    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Sources.Batteries.Lithium_Ion_LFP
    """

    # ---------------------------------------------------------------------------------    
    # battery cell properties
    # --------------------------------------------------------------------------------- 
    electrode_area            = battery_module.cell.electrode_area 
    As_cell                   = battery_module.cell.surface_area
    cell_mass                 = battery_module.cell.mass    
    Cp                        = battery_module.cell.specific_heat_capacity       
    battery_module_data       = battery_module.cell.discharge_performance_map
    
    # ---------------------------------------------------------------------------------
    # Compute Bus electrical properties 
    # ---------------------------------------------------------------------------------    
    bus_conditions              = state.conditions.energy[bus.tag]
    bus_config                  = bus.battery_module_electric_configuration
    phi                         = state.conditions.energy.hybrid_power_split_ratio
    psi                         = state.conditions.energy.battery_fuel_cell_power_split_ratio
    E_bus                       = bus_conditions.energy
    P_bus                       = bus_conditions.power_draw*phi * psi
    I_bus                       = bus_conditions.current_draw*phi * psi
    
    # ---------------------------------------------------------------------------------
    # Compute battery_module Conditions
    # -------------------------------------------------------------------------    
    battery_module_conditions = state.conditions.energy[bus.tag].battery_modules[battery_module.tag]  
   
    E_module_max       = battery_module.maximum_energy * battery_module_conditions.cell.capacity_fade_factor
    
    V_oc_module        = battery_module_conditions.voltage_open_circuit
    V_oc_cell          = battery_module_conditions.cell.voltage_open_circuit   
  
    P_module           = battery_module_conditions.power
    P_cell             = battery_module_conditions.cell.power
    
    R_0_module         = battery_module_conditions.internal_resistance
    R_0_cell           = battery_module_conditions.cell.internal_resistance
    
    Q_heat_module      = battery_module_conditions.heat_energy_generated
    Q_heat_cell        = battery_module_conditions.cell.heat_energy_generated
    
    V_ul_module        = battery_module_conditions.voltage_under_load
    V_ul_cell          = battery_module_conditions.cell.voltage_under_load
    
    I_module           = battery_module_conditions.current 
    I_cell             = battery_module_conditions.cell.current
    
    T_module           = battery_module_conditions.temperature                 
    T_cell             = battery_module_conditions.cell.temperature
    
    SOC_cell           = battery_module_conditions.cell.state_of_charge  
    SOC_module         = battery_module_conditions.state_of_charge
    E_cell             = battery_module_conditions.cell.energy   
    E_module           = battery_module_conditions.energy
    Q_cell             = battery_module_conditions.cell.charge_throughput              
    DOD_cell           = battery_module_conditions.cell.depth_of_discharge
    
    # ---------------------------------------------------------------------------------
    # Compute battery_module electrical properties 
    # -------------------------------------------------------------------------    
    # Calculate the current going into one cell  
    n_series          = battery_module.electrical_configuration.series
    n_parallel        = battery_module.electrical_configuration.parallel 
    n_total           = n_series*n_parallel 
    no_modules        = len(bus.battery_modules)
    
    # ---------------------------------------------------------------------------------
    # Examine Thermal Management System
    # ---------------------------------------------------------------------------------
    HAS = None  
    for coolant_line in coolant_lines:
        for tag, item in  coolant_line.items():
            if tag == 'battery_modules':
                for sub_tag, sub_item in item.items():
                    if sub_tag == battery_module.tag:
                        for btms in  sub_item:
                            HAS = btms    


    # ---------------------------------------------------------------------------------------------------
    # Current State 
    # ---------------------------------------------------------------------------------------------------
    if bus_config == 'Series':
        I_module[t_idx]      = I_bus[t_idx]
    elif bus_config  == 'Parallel':
        I_module[t_idx]      = I_bus[t_idx] /len(bus.battery_modules)

    I_cell[t_idx] = I_module[t_idx] / n_parallel   
       
    # ---------------------------------------------------------------------------------
    # Compute battery_module cell temperature 
    # ---------------------------------------------------------------------------------
    R_0_cell[t_idx]                     =  (0.01483*(SOC_cell[t_idx]**2) - 0.02518*SOC_cell[t_idx] + 0.1036) *battery_module_conditions.cell.resistance_growth_factor  
    R_0_cell[t_idx][R_0_cell[t_idx]<0]  = 0. 

    # Determine temperature increase         
    sigma                 = 139 # Electrical conductivity
    n                     = 1
    F                     = 96485 # C/mol Faraday constant    
    delta_S               = -496.66*(SOC_cell[t_idx])**6 +  1729.4*(SOC_cell[t_idx])**5 + -2278 *(SOC_cell[t_idx])**4 +  1382.2 *(SOC_cell[t_idx])**3 + \
                            -380.47*(SOC_cell[t_idx])**2 +  46.508*(SOC_cell[t_idx])  + -10.692  

    i_cell                = I_cell[t_idx]/electrode_area # current intensity
    q_dot_entropy         = -(T_cell[t_idx])*delta_S*i_cell/(n*F)       
    q_dot_joule           = (i_cell**2)*(battery_module_conditions.cell.resistance_growth_factor)/(sigma)          
    Q_heat_cell[t_idx]    = (q_dot_joule + q_dot_entropy)*As_cell 
    Q_heat_module[t_idx]  = Q_heat_cell[t_idx]*n_total  

    V_ul_cell[t_idx]      = compute_nmc_cell_state(battery_module_data,SOC_cell[t_idx],T_cell[t_idx],abs(I_cell[t_idx])) 

    V_oc_cell[t_idx]      = V_ul_cell[t_idx] + (abs(I_cell[t_idx]) * R_0_cell[t_idx])              

    # Effective Power flowing through battery_module 
    P_module[t_idx]       = P_bus[t_idx] /no_modules  - np.abs(Q_heat_module[t_idx]) 

    # store remaining variables 
    V_oc_module[t_idx]     = V_oc_cell[t_idx]*n_series 
    V_ul_module[t_idx]     = V_ul_cell[t_idx]*n_series  
    T_module[t_idx]        = T_cell[t_idx]   # Assume the cell temperature is the temperature of the module
    P_cell[t_idx]          = P_module[t_idx]/n_total 
    E_module[t_idx]        = E_bus[t_idx]/no_modules 
    E_cell[t_idx]          = E_module[t_idx]/n_total  

    # ---------------------------------------------------------------------------------------------------     
    # Future State 
    # --------------------------------------------------------------------------------------------------- 
    if t_idx != state.numerics.number_of_control_points-1:  

        # Compute cell temperature
        if HAS is not None:
            T_cell[t_idx+1]  = HAS.compute_thermal_performance(battery_module,bus,coolant_line,Q_heat_cell[t_idx],T_cell[t_idx],state,delta_t[t_idx],t_idx)
        else:
            # Considers a thermally insulated system and the heat piles on in the system
            dT_dt              = Q_heat_cell[t_idx]/(cell_mass*Cp)
            T_cell[t_idx+1]    =  T_cell[t_idx] + dT_dt*delta_t[t_idx]
            
        # Compute state of charge and depth of discarge of the battery_module
        E_module[t_idx+1]                                     = (E_module[t_idx]) -P_module[t_idx]*delta_t[t_idx]
        E_module[t_idx+1][E_module[t_idx+1] > E_module_max]   = np.float32(E_module_max)
        SOC_cell[t_idx+1]                                     = E_module[t_idx+1]/E_module_max 
        SOC_cell[t_idx+1][SOC_cell[t_idx+1]>1]                = 1.
        SOC_cell[t_idx+1][SOC_cell[t_idx+1]<0]                = 0. 
        DOD_cell[t_idx+1]                                     = 1 - SOC_cell[t_idx+1]  
        SOC_module[t_idx+1]                                   = SOC_cell[t_idx+1]

    
        # Determine new charge throughput (the amount of charge gone through the battery_module)
        Q_cell[t_idx+1]    = Q_cell[t_idx] + abs(I_cell[t_idx])*delta_t[t_idx]/Units.hr
        
    stored_results_flag     = True
    stored_battery_module_tag     = battery_module.tag  
        
    return stored_results_flag, stored_battery_module_tag


def reuse_stored_nmc_cell_data(battery_module,state,bus,stored_results_flag, stored_battery_module_tag):
    '''Reuses results from one propulsor for identical batteries
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    

    Outputs:  
    
    Properties Used: 
    N.A.        
    '''
   
    state.conditions.energy[bus.tag].battery_modules[battery_module.tag] = deepcopy(state.conditions.energy[bus.tag].battery_modules[stored_battery_module_tag])
    
        
    return
 
def compute_nmc_cell_state(battery_module_data, SOC, T, I):
    """
    Computes the electrical state variables of a lithium-nickel-manganese-cobalt-oxide (NMC) battery cell using look-up tables.
    
    Parameters
    ----------
    battery_module_data : function
        Look-up function that maps state of charge, temperature, and current to voltage
    SOC : numpy.ndarray
        State of charge of the cell [unitless, 0-1]
    T : numpy.ndarray
        Battery cell temperature [K]
    I : numpy.ndarray
        Battery cell current [A]
    
    Returns
    -------
    V_ul : numpy.ndarray
        Under-load voltage [V]
    
    Notes
    -----
    This function computes the voltage of an NMC battery cell under load conditions
    by using a look-up table approach. It converts the state of charge to depth of discharge,
    and then uses this value along with temperature and current to determine the cell voltage.
    
    The function applies limits to ensure the inputs are within the valid range of the
    look-up data:
        - SOC is limited to [0, 1]
        - Temperature is limited to [272.65K, 322.65K] (approximately 0°C to 50°C)
        - Current is limited to [0A, 8A]
    
    The input to the look-up table is a concatenated array of [current, temperature, depth_of_discharge].
    
    **Major Assumptions**
        * The battery performance can be accurately represented by a look-up table
        * The model is valid only within the specified temperature and current ranges
    
    See Also
    --------
    RCAIDE.Library.Components.Powertrain.Sources.Battery_Modules.Lithium_Ion_NMC
    """

    # Make sure things do not break by limiting current, temperature and current 
    SOC[SOC < 0.]   = 0.  
    SOC[SOC > 1.]   = 1.    
    DOD             = 1 - SOC 
    
    T[np.isnan(T)] = 302.65
    T[T<272.65]    = 272.65 # model does not fit for below 0  degrees
    T[T>322.65]    = 322.65 # model does not fit for above 50 degrees
     
    I[I<0.0]       = 0.0
    I[I>8.0]       = 8.0   
     
    pts            = np.hstack((np.hstack((I, T)),DOD  )) # amps, temp, SOC   
    V_ul           = np.atleast_2d(battery_module_data.Voltage(pts)[:,1]).T  
    
    return V_ul


# RCAIDE/Library/Missions/Segments/Descent/Constant_EAS_Constant_Rate.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------  
#  IMPORT 
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports 
import RCAIDE

# pacakge imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------  
def initialize_conditions(segment):
    """
    Initializes conditions for constant equivalent airspeed descent at fixed rate

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed
            - descent_rate : float
                Rate of descent [m/s]
            - equivalent_air_speed : float
                Equivalent airspeed to maintain [m/s]
            - altitude_start : float
                Initial altitude [m]
            - altitude_end : float
                Final altitude [m]
            - sideslip_angle : float
                Aircraft sideslip angle [rad]
            - state:
                numerics.dimensionless.control_points : array
                    Discretization points [-]
                conditions : Data
                    State conditions container
            - analyses:
                atmosphere : Model
                    Atmospheric model for property calculations
    
    Returns
    -------
    None

    Notes
    -----
    This function sets up the initial conditions for a descent segment with constant
    equivalent airspeed (EAS) and constant descent rate. It handles the conversion
    between EAS and true airspeed accounting for density variations with altitude.
    Updates segment conditions directly with velocity_vector [m/s], altitude [m],
    and position_vector [m].

    **Calculation Process**
        1. Discretize altitude profile
        2. Get atmospheric properties at each altitude
        3. Convert EAS to true airspeed:
            V = EAS/sqrt(ρ/ρ0) where:
                - ρ is local density
                - ρ0 is sea level density
        4. Decompose velocity into components using:
            - Fixed descent rate
            - Sideslip angle
            - Computed true airspeed

    **Major Assumptions**
        * Constant equivalent airspeed
        * Constant descent rate
        * Standard atmosphere model
        * Small angle approximations
        * Quasi-steady flight

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    RCAIDE.Library.Mission.Common.Update.atmosphere
    """       
    
    # unpack
    descent_rate = segment.descent_rate
    eas          = segment.equivalent_air_speed   
    alt0         = segment.altitude_start 
    altf         = segment.altitude_end 
    beta         = segment.sideslip_angle
    t_nondim     = segment.state.numerics.dimensionless.control_points
    conditions   = segment.state.conditions  

    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]


    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0
    
    # Pack altitude before updating atmosphere
    conditions.freestream.altitude[:,0]             =  alt[:,0] # positive altitude in this context
    
    # determine airspeed from equivalent airspeed
    RCAIDE.Library.Mission.Common.Update.atmosphere(segment) # get density for airspeed
    density   = conditions.freestream.density[:,0]   
    MSL_data  = segment.analyses.atmosphere.compute_values(0.0,0.0)

    # check for initial velocity vector
    if eas is None:
        if not segment.state.initials: raise AttributeError('initial equivalent airspeed not set')
        air_speed  = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1,:])  
    else: 
        air_speed  = eas/np.sqrt(density/MSL_data.density[0])    
    
    # process velocity vector
    v_mag = air_speed
    v_z   = descent_rate # z points down
    v_xy  = np.sqrt( v_mag**2 - v_z**2 )
    v_x   = np.cos(beta)*v_xy
    v_y   = np.sin(beta)*v_xy
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0] = v_x
    conditions.frames.inertial.velocity_vector[:,1] = v_y
    conditions.frames.inertial.velocity_vector[:,2] = v_z
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down



# RCAIDE/Library/Missions/Common/Residuals/flight_dynamics.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

import RCAIDE
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  Residual Total Forces
# ---------------------------------------------------------------------------------------------------------------------- 
def flight_dynamics(segment):
    """
    Evaluates flight dynamics residuals for mission segment analysis

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed

    Returns
    -------
    None
        Updates segment residuals directly

    Notes
    -----
    This function calculates the residuals for force and moment equations
    in all three axes. It handles special cases for transition and ground
    segments, including acceleration calculations and final velocity constraints.

    The function processes:
        1. Force equation residuals (F = ma)
        2. Moment equation residuals (M = Iα)
        3. Special handling for:
            - Transition segments
            - Ground operations (takeoff, landing)

    **Required Segment State Variables**

    state.conditions.frames.inertial:
        - velocity_vector : array
            Vehicle velocity [m/s]
        - acceleration_vector : array
            Vehicle acceleration [m/s²]
        - total_force_vector : array
            Net forces [N]
        - total_moment_vector : array
            Net moments [N⋅m]
        - angular_velocity_vector : array
            Angular rates [rad/s]
        - angular_acceleration_vector : array
            Angular accelerations [rad/s²]

    state.conditions.weights:
        - total_mass : array
            Vehicle mass [kg]

    analyses.aerodynamics.vehicle.mass_properties:
        - moments_of_inertia.tensor : array
            Inertia tensor [kg⋅m²]

    **Segment Types**
    
    Special handling for:
    - Transition segments
        * Constant acceleration
        * Constant angle
        * Linear climb
    - Ground segments
        * Takeoff
        * Landing
        * Ground operations

    **Major Assumptions**
        * Rigid body dynamics
        * Principal axes aligned with body axes
        * Constant inertia properties
        * Valid mass properties
        * Non-zero final velocity for ground segments

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    RCAIDE.Framework.Mission.Segments.Transition
    RCAIDE.Framework.Mission.Segments.Ground
    """
    transition_seg_flag =  type(segment) == RCAIDE.Framework.Mission.Segments.Transition.Constant_Acceleration_Constant_Angle_Linear_Climb
    ground_seg_flag =  (type(segment) == RCAIDE.Framework.Mission.Segments.Ground.Landing) or\
        (type(segment) == RCAIDE.Framework.Mission.Segments.Ground.Takeoff) or \
        (type(segment) == RCAIDE.Framework.Mission.Segments.Ground.Ground)

    if transition_seg_flag or ground_seg_flag: 
        v       = segment.state.conditions.frames.inertial.velocity_vector
        D       = segment.state.numerics.time.differentiate 
        segment.state.conditions.frames.inertial.acceleration_vector = np.dot(D,v)

    FT = segment.state.conditions.frames.inertial.total_force_vector
    a  = segment.state.conditions.frames.inertial.acceleration_vector    

    if transition_seg_flag: 
        omega = segment.state.conditions.frames.inertial.angular_velocity_vector
        D  = segment.state.numerics.time.differentiate   
        segment.state.conditions.frames.inertial.angular_acceleration_vector =  np.dot(D,omega)         

    MT      = segment.state.conditions.frames.inertial.total_moment_vector    
    ang_acc = segment.state.conditions.frames.inertial.angular_acceleration_vector  
    m       = segment.state.conditions.weights.total_mass
    I       = segment.analyses.aerodynamics.vehicle.mass_properties.moments_of_inertia.tensor  

    if ground_seg_flag:
        vf = segment.velocity_end
        if vf == 0.0: vf = 0.01 
        segment.state.residuals.force_x[:,0] = FT[1:,0]/m[1:,0] - a[1:,0] 
        segment.state.residuals.final_velocity_error = (v[-1,0] - vf)
    else: 
        if segment.flight_dynamics.force_x: 
            segment.state.residuals.force_x[:,0] = FT[:,0]/m[:,0] - a[:,0]  
        if segment.flight_dynamics.force_y: 
            segment.state.residuals.force_y[:,0] = FT[:,1]/m[:,0] - a[:,1]    
        if segment.flight_dynamics.force_z: 
            segment.state.residuals.force_z[:,0] = FT[:,2]/m[:,0] - a[:,2]  
        if  segment.flight_dynamics.moment_x:
            segment.state.residuals.moment_x[:,0] = MT[:,0]/I[0,0] - ang_acc[:,0]   
        if  segment.flight_dynamics.moment_y:
            segment.state.residuals.moment_y[:,0] = MT[:,1]/I[1,1] - ang_acc[:,1]   
        if  segment.flight_dynamics.moment_z:
            segment.state.residuals.moment_z[:,0] = MT[:,2]/I[2,2] - ang_acc[:,2] 
     
    return



# RCAIDE/Library/Plots/Noise/plot_3D_noise_contour.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots  import *

# python imports 
import numpy as np  
import plotly.graph_objects as go

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------    
def plot_3D_noise_contour(noise_data,
                       noise_level = None,
                       min_noise_level = 35,  
                       max_noise_level = 90, 
                       noise_scale_label = None,
                       save_figure = False,
                       show_figure = True,
                       save_filename = "Noise_Contour",
                       use_lat_long_coordinates = True, 
                       show_trajectory = False,
                       show_microphones = False,
                       colormap = 'jet',
                       file_type = ".png",
                       background_color = 'rgb(17,54,71)',
                       grid_color = 'gray',
                       width = 1400, 
                       height = 800):
    """
    Creates an interactive 3D visualization of noise contours with optional aircraft trajectory.

    Parameters
    ----------
    noise_data : NoiseData
        RCAIDE noise data structure containing:
            - microphone_locations[:,:,0:3]
                3D array of microphone positions in (nmi, nmi, ft)
            - aircraft_position[:,0:3]
                Aircraft trajectory points in (nmi, nmi, ft)
            
    noise_level : ndarray
        2D array of noise levels at measurement points
        
    min_noise_level : float, optional
        Minimum noise level for contour scale (default: 35 dB)
        
    max_noise_level : float, optional
        Maximum noise level for contour scale (default: 90 dB)
        
    noise_scale_label : str, optional
        Label for noise metric (e.g., "dBA", "EPNL", etc.)
        
    save_figure : bool, optional
        Flag for saving the figure (default: False)
        
    show_figure : bool, optional
        Flag to display the figure (default: True)
        
    save_filename : str, optional
        Name of file for saved figure (default: "Noise_Contour")
        
    use_lat_long_coordinates : bool, optional
        Flag to use geographic coordinates (default: True)
        
    show_trajectory : bool, optional
        Flag to display aircraft trajectory (default: False)
        
    show_microphones : bool, optional
        Flag to display microphone locations (default: False)
        
    colormap : str, optional
        Colormap specification for noise contours (default: 'jet')
        
    file_type : str, optional
        File extension for saved figure (default: ".png")
        
    background_color : str, optional
        Color specification for plot background (default: 'rgb(17,54,71)')
        
    grid_color : str, optional
        Color specification for grid lines (default: 'gray')
        
    width : int, optional
        Figure width in pixels (default: 1400)
        
    height : int, optional
        Figure height in pixels (default: 800)

    Returns
    -------
    fig_3d : plotly.graph_objects.Figure
        Handle to the generated interactive 3D figure

    Notes
    -----
    Creates visualization showing:
        * 3D noise contour surface
        * Optional aircraft trajectory
        * Optional microphone locations
        * Interactive viewing controls
        * Customizable appearance
    
    **Major Assumptions**
        * Noise levels are in decibels
        * Coordinates are in nautical miles and feet
        * Measurement grid is regularly spaced
        * Z-axis represents elevation
    
    **Definitions**
    
    'Noise Contour'
        Surface of constant noise level
    'Aircraft Trajectory'
        Time history of aircraft position
    'Microphone Location'
        Measurement point coordinates
    
    See Also
    --------
    RCAIDE.Library.Plots.Noise.plot_2D_noise_contour : 2D visualization of noise field
    RCAIDE.Library.Plots.Noise.contour_surface_slice : Surface generation utility
    """   
    Aircraft_pos    = noise_data.aircraft_position      
    X               = noise_data.microphone_locations[:,:,0]/Units.nmi  
    Y               = noise_data.microphone_locations[:,:,1]/Units.nmi  
    Z               = noise_data.microphone_locations[:,:,2]/Units.feet  
    plot_data       = []   
  
    # ---------------------------------------------------------------------------
    # TRHEE DIMENSIONAL NOISE CONTOUR
    # --------------------------------------------------------------------------- 
    # TERRAIN CONTOUR 
    ground_contour   = contour_surface_slice(Y,X,Z,noise_level,color_scale=colormap)
    plot_data.append(ground_contour)

    # GROUND MICROPHONES
    if show_microphones:
        microphones = go.Scatter3d(x        = Y.flatten(),
                                   y        = X.flatten(),
                                   z        = Z.flatten(),
                                   mode     = 'markers',
                                   marker   = dict(size=6,color='white',opacity=0.8),
                                   line     = dict(color='white',width=2))
        plot_data.append(microphones)

    # AIRCRAFT TRAJECTORY
    if show_trajectory:
        aircraft_trajectory = go.Scatter3d(x   = Aircraft_pos[:,1]/Units.nmi,
                                           y   = Aircraft_pos[:,0]/Units.nmi,
                                           z   = Aircraft_pos[:,2]/Units.feet,
                                           mode= 'markers',
                                           marker=dict(size=6,
                                                       color='black',
                                                       opacity=0.8),
                                    line=dict(color='black',width=2))
        plot_data.append(aircraft_trajectory)

    # Define Colorbar Bounds
    min_alt     = np.minimum(np.min(Z),0) 
    max_alt     = np.maximum(np.max(Z), np.max(Aircraft_pos[:,2]/Units.feet)) 
  
    fig_3d = go.Figure(data=plot_data) 

    if show_microphones or show_trajectory:
        pass
    else: 
        fig_3d.update_traces(colorbar_orientation     = 'v',
                             colorbar_thickness       = 50,
                             colorbar_nticks          = 10,
                             colorbar_title_text      = noise_scale_label,
                             colorbar_tickfont_size   = 16,
                             colorbar_title_side      = "right",
                             colorbar_ypad            = 50,
                             colorbar_len             = 0.75)
        
                         
    fig_3d.update_layout(
             title_text                             = save_filename, 
             title_x                                = 0.5,
             width                                  = width,
             height                                 = height,
             font_size                              = 12,
             scene_aspectmode                       = "manual",
             scene_aspectratio                      = dict(x=1, y=1, z=0.5),      
             scene_zaxis_range                      = [min_alt,max_alt*3],
             scene                                  = dict(xaxis_title='Latitude [nmi]',
                                                           yaxis_title='Longitude [nmi]',
                                                           zaxis_title='Elevation [ft]',
                                                           xaxis = dict(
                                                                backgroundcolor=background_color,
                                                                gridcolor="white",
                                                                showbackground=True,
                                                                zerolinecolor=grid_color,),
                                                           yaxis = dict(
                                                               backgroundcolor=background_color,
                                                               gridcolor=grid_color,
                                                               showbackground=True,
                                                               zerolinecolor="white"),
                                                           zaxis = dict(
                                                               backgroundcolor=background_color,
                                                               gridcolor=grid_color,
                                                               showbackground=True,
                                                               zerolinecolor="white",),),
             scene_camera=dict(up    = dict(x=0, y=0, z=1),
                               center= dict(x=-0.05, y=0, z=-0.20),
                               eye   = dict(x=-1.0, y=-1.0, z=.4))   
    ) 
    if show_figure:
        fig_3d.show() 
    if save_figure:
        fig_3d.write_image(save_filename, file_type)

    return fig_3d       

def colorax(vmin, vmax):
    return dict(cmin=vmin, cmax=vmax)
 


# RCAIDE/Library/Methods/Powertrain/Converters/Engine/append_engine_conditions.py
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_engine_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_engine_conditions(engine,segment,energy_conditions,noise_conditions=None): 
    """
    Initializes and appends engine operating conditions to the propulsor conditions data structure.
    
    Parameters
    ----------
    engine : RCAIDE.Library.Components.Propulsors
        Engine system instance for which conditions are being initialized
    segment : RCAIDE.Framework.Mission.Segments.Segment
        Mission segment instance containing flight conditions
    energy_conditions : dict
        variable onto which engine operating conditions are appended
    noise_conditions : dict, optional
        variable onto which engine noise conditions are appended. Unused for now.
        
    Returns
    -------
    None
        
    Notes
    -----
    This function creates a nested structure of Conditions objects to store engine
    inputs and outputs during mission analysis. The conditions are stored under
    the engine's unique tag identifier.
    """
    # unpack 
    ones_row          = segment.state.ones_row
    
    energy_conditions.converters[engine.tag]                      = Conditions() 
    energy_conditions.converters[engine.tag].inputs               = Conditions()
    energy_conditions.converters[engine.tag].outputs              = Conditions()
    energy_conditions.converters[engine.tag].omega                = engine.rated_speed * ones_row(1)   
    
    return 
